# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""

Minimum cost design of end-plate connections

@author: Kristo Mela
@date: 31.1.2020
"""

import numpy as np
import math
import time

MIN_HEIGHT = 100
MAX_HEIGHT = 500
step_height = 10
MIN_WIDTH = 100
MAX_WIDTH = 500
step_width = 10

HEIGHTS = np.arange(MIN_HEIGHT, MAX_HEIGHT+step_height, step_height)
WIDTHS = np.arange(MIN_WIDTH, MAX_WIDTH+step_width, step_width)
#THICKNESSES = [5, 6, 8, 10, 12, 14, 15, 16, 18, 20, 22, 25, 30, 35, 40, 50]
THICKNESSES = [4, 5, 6, 8, 10, 12, 14, 15, 16, 18, 20]
MAX_THICK = max(THICKNESSES)
MIN_THICK = min(THICKNESSES)

try:
    from metku.frame2d.frame2d import *
    import metku.frame2d.frame2d as f2d
    from metku.optimization.structopt import *
    # import metku.optimization.structopt as sopt
    from metku.optimization.solvers.trust_region import TrustRegionConstr
    from metku.optimization.solvers.bnb import BnB
    from metku.optimization.solvers.lp import LP
except:
    
    from optimization.structopt import *
    from optimization.solvers.trust_region import TrustRegionConstr    
        
class EndPlateJointOpt(OptimizationProblem):
    """
        
    """
    def __init__(self,end_plate_joint):
        super().__init__("End plate connection")

        self.prob_type = prob_type
        self.cons.clear()        
        self.clear_vars()
        self.structure = None        
        self.LT_buckling = LT_buckling
        self.buckling_z = buckling_z
        self.symmetry = symmetry
        self.top_flange_class = top_flange_class
        self.bottom_flange_class = bottom_flange_class
        self.web_class = web_class
        self.L = L
        self.Lpi = Lpi
        self.Fx = Fx
        self.Fy = Fy
        self.Qx = Qx
        self.lcr = lcr
        
        section_class = max(top_flange_class, bottom_flange_class)
        self.create_structure(Lpi, Fx, Fy, Qx, Qy, Mz, lcr, LT_buckling)
        self.create_variables()
        self.create_constraints(section_class)
        self.create_objective()
        self.prob_type = 'continuous'

    def create_objective(self):
        def obj_fun(x):
            self.substitute_variables(x)
            return self.structure.weight
        obj = ObjectiveFunction("weight", obj_fun,problem=self)
        self.add(obj)

    def create_structure(self, Lpi, Fx, Fy, Qx, Qy, Mz, lcr, LT_buckling):
        # Luo tyhjän kehän
        frame = Frame2D(num_elements=4)
        # Luo pilarin (koordinaatit, profile=vapaaehtoinen)
        col = SteelColumn([[0, 0], [0, Lpi]], LT_buckling,
                          profile='WI 500-12-10X300-10X300',
                          material='S355MC')
        # Lisätään pilarille materiaali
        col.material = "S355MC"
        # Lisätään nurjahduspituus (masto lcr[0]=2, muuten lcr[0]=0.7)
        col.steel_member.lcr[0] = lcr
        # Lisää pilarin kehälle
        # col.profile = 'WI 800-12-30X450-25X300'
        frame.add(col)
        # Lisää niveltuen pisteeseen (0,Lpi)
        if lcr == 0.7:
            frame.add(XHingedSupport([0, Lpi]))
        # Lisää jäykän tuen pisteeseen (0,0)
        frame.add(FixedSupport([0, 0]))
        # Lisää pistekuorman pilarin yläpäähän
        frame.add(f2d.PointLoad([0, Lpi], [Fx, Fy, Mz]))
        # Lisää tasaisen kuorman rakenneosalle
        frame.add(f2d.LineLoad(col, [Qx, Qx], "x"))
        # Luo kehän fem -mallin
        frame.generate()
        # Laskee
        frame.calculate()
        # Piirtää kehän (pilarin) näytölle
        # frame.plot()

        self.structure = frame

    def create_variables(self):
        """
        Creates design variables
        """

        col = self.structure.members[0].cross_section

        if self.prob_type == "continuous":

            var_h = Variable("h", MIN_HEIGHT, MAX_HEIGHT,
                             target={"property": "h", "objects": [col]})
            var_tw = Variable("tw", MIN_THICK, MAX_THICK,
                              target={"property": "tw", "objects": [col]})
            if self.symmetry == "mono":
                var_tt = Variable("tt", MIN_THICK, MAX_THICK,
                                 target={"property": "tt", "objects": [col]})
                var_tb = Variable("tb", MIN_THICK, MAX_THICK,
                                 target={"property": "tb", "objects": [col]})
                var_bt = Variable("bt", MIN_WIDTH, MAX_WIDTH,
                                  target={"property": "bt", "objects": [col]})
                var_bb = Variable("bb", MIN_WIDTH, MAX_WIDTH,
                                  target={"property": "bb", "objects": [col]})

                vars = [var_h, var_tw, var_bt, var_tt, var_bb, var_tb]
                for var in vars:
                    self.add(var)

            elif self.symmetry == "dual":
                var_tf = Variable("tf", MIN_THICK, MAX_THICK,
                                  target={"property": "tf", "objects": [col]})
                var_bf = Variable("b", MIN_WIDTH, MAX_WIDTH,
                                  target={"property": "b", "objects": [col]})

                vars = [var_h, var_tw, var_bf, var_tf]
                for var in vars:
                    self.add(var)

            else:
                raise ValueError("Symmetry must be either dual or mono")

        elif self.prob_type == "discrete":
            var_h = DiscreteVariable(
                "h", values=HEIGHTS,
                target={"property": "h", "objects": [col]})
            var_tw = DiscreteVariable(
                "tw", values=THICKNESSES,
                target={"property": "tw", "objects": [col]})

            var_tw.branch_priority = 1
            if self.symmetry == "mono":
                var_tt = DiscreteVariable(
                    "tt", values=THICKNESSES,
                    target={"property": "tt", "objects": [col]})
                var_tb = DiscreteVariable(
                    "tb", values=THICKNESSES,
                    target={"property": "tb", "objects": [col]})
                var_bt = DiscreteVariable(
                    "bt", values=WIDTHS,
                    target={"property": "bt", "objects": [col]})
                var_bb = DiscreteVariable(
                    "bb", values=WIDTHS,
                    target={"property": "bb", "objects": [col]})

                vars = [var_h, var_tw, var_bt, var_tt, var_bb, var_tb]
                for var in vars:
                    self.add(var)

            elif self.symmetry == "dual":
                var_tf = DiscreteVariable(
                    "tf", values=THICKNESSES,
                    target={"property": "tf", "objects": [col]})
                
                var_tf.branch_priority = 1
                var_bf = DiscreteVariable(
                    "b", values=WIDTHS,
                    target={"property": "b", "objects": [col]})

                vars = [var_h, var_tw, var_bf, var_tf]
                for var in vars:
                    self.add(var)

            else:
                raise ValueError("Symmetry must be either dual or mono")

        elif self.prob_type == "index":
            var_h = IndexVariable(
                "h", values=HEIGHTS,
                target={"property": "h", "objects": [col]})
            var_tw = IndexVariable(
                "tw", values=THICKNESSES,
                target={"property": "tw", "objects": [col]})

            if self.symmetry == "mono":
                var_tt = IndexVariable(
                    "tt", values=THICKNESSES,
                    target={"property": "tt", "objects": [col]})
                var_tb = IndexVariable(
                    "tb", values=THICKNESSES,
                    target={"property": "tb", "objects": [col]})
                var_bt = IndexVariable(
                    "bt", values=WIDTHS,
                    target={"property": "bt", "objects": [col]})
                var_bb = IndexVariable(
                    "bb", values=WIDTHS,
                    target={"property": "bb", "objects": [col]})

                vars = [var_h, var_tw, var_bt, var_tt, var_bb, var_tb]
                for var in vars:
                    self.add(var)

            elif self.symmetry == "dual":
                var_tf = IndexVariable(
                    "tf", values=THICKNESSES,
                    target={"property": "tf", "objects": [col]})
                var_bf = IndexVariable(
                    "b", values=WIDTHS,
                    target={"property": "b", "objects": [col]})

                vars = [var_h, var_tw, var_bf, var_tf]
                for var in vars:
                    self.add(var)

            else:
                raise ValueError("Symmetry must be either dual or mono")

    def WIColumnTopFlangeClassCon(self, mem):
        """
        Builds a constraint for cross-section class of a WI column
        0.5*bt - 0.5*tw - C0*e*tt <= sqrt(2)*aw
        """

        #a = np.zeros_like(self.vars)
        a = np.zeros(self.nvars())
        con_type = '<'

        e = mem.cross_section.eps

        if self.top_flange_class == 1:
            C0 = 9
        elif self.top_flange_class == 2:
            C0 = 10
        elif self.top_flange_class > 2:
            C0 = 14

        for i in range(len(self.vars)):

            if self.vars[i].target["property"] == "bt" or \
                    self.vars[i].target["property"] == "b":
                a[i] = 0.5
            elif self.vars[i].target["property"] == "tt" or \
                    self.vars[i].target["property"] == "tf":
                a[i] = - C0 * e
            elif self.vars[i].target["property"] == "tw":
                a[i] = - 0.5

        b = math.sqrt(2) * mem.cross_section.weld_throat

        # print(a)

        if self.top_flange_class > 3:
            """ if class 4 is required, the cf/tf ratio needs to
                be greater than the class 3 limit. This  changes the
                direction of the constraint from < to >.
            """
            con_type = '>'

        con_name = "Top flange in class " + str(self.top_flange_class)
        con = LinearConstraint(a, b, con_type, name=con_name)

        return con

    def WIColumnBottomFlangeClassCon(self, mem):
        """
        Builds a constraint for cross-section class of a WI column
        0.5*bt - 0.5*tw - C0*e*tt <= sqrt(2)*aw
        """

        #a = np.zeros_like(self.vars)
        a = np.zeros(self.nvars())
        con_type = '<'

        e = mem.cross_section.eps

        if self.bottom_flange_class == 1:
            C0 = 9
        elif self.bottom_flange_class == 2:
            C0 = 10
        elif self.bottom_flange_class > 2:
            C0 = 14

        for i in range(len(self.vars)):

            if self.vars[i].target["property"] == "bt":
                a[i] = 0.5
            elif self.vars[i].target["property"] == "tt":
                a[i] = - C0 * e
            elif self.vars[i].target["property"] == "tw":
                a[i] = - 0.5

        b = math.sqrt(2) * mem.cross_section.weld_throat

        # print(a)

        if self.top_flange_class > 3:
            """ if class 4 is required, the cf/tf ratio needs to
                be greater than the class 3 limit. This  changes the
                direction of the constraint from < to >.
            """
            con_type = '>'

        con_name = "Bottom flange in class " + str(self.bottom_flange_class)
        con = LinearConstraint(a, b, con_type, name=con_name)

        return con

    def WIColumnWebClassCon(self, mem):
        """
        Builds a constraint for cross-section class of a WI column
        h - tt - tb - C1*e*tw <= 2*sqrt(2)*aw
        """

        #a = np.zeros_like(self.vars)
        a = np.zeros(self.nvars())
        con_type = '<'

        e = mem.cross_section.eps

        # web is assumed to be in bending
        # if self.web_class == 1:
        #     C1 = 72
        # elif self.web_class == 2:
        #     C1 = 83
        # elif self.web_class > 2:
        #     C1 = 124

        # web is assumed to be in compression
        if self.web_class == 1:
            C1 = 33
        elif self.web_class == 2:
            C1 = 38
        elif self.web_class > 2:
            C1 = 42

        for i in range(len(self.vars)):

            if self.vars[i].target["property"] == "h":
                a[i] = 1
            elif self.vars[i].target["property"] == "tt":
                a[i] = - 1
            elif self.vars[i].target["property"] == "tb":
                a[i] = - 1
            elif self.vars[i].target["property"] == "tw":
                a[i] = - C1 * e
            elif self.vars[i].target["property"] == "tf":
                a[i] = -2

        b = 2 * math.sqrt(2) * mem.cross_section.weld_throat

        # print("PRINT", a)

        if self.web_class > 3:
            """ if class 4 is required, the cw/tw ratio needs to
                be greater than the class 3 limit. This  changes the
                direction of the constraint from < to >.
            """
            con_type = '>'

        con_name = "Web in class " + str(self.web_class)
        con = LinearConstraint(a, b, con_type, name=con_name)

        return con

    def WIColumnWebHeightCon(self, h_min=50):
        """
        Constraint for ensuring that the web height is positive, at least 'h_min'
        h-tt-tb >= h_min, or
        -h + tt + tb <= -h_min
        """

        #a = np.zeros_like(self.vars)
        a = np.zeros(self.nvars())

        con_type = '<'

        for i in range(len(self.vars)):

            if self.vars[i].target["property"] == "h":
                a[i] = -1
            elif self.vars[i].target["property"] == "tt":
                a[i] = + 1
            elif self.vars[i].target["property"] == "tb":
                a[i] = + 1            
            elif self.vars[i].target["property"] == "tf":
                a[i] = +2

        b = -h_min

        con_name = "Web height at least " + str(h_min)

        con = LinearConstraint(a, b, con_type, name=con_name)

        return con

    def TopFlangeShearLagCon(self, mem):
        """
        Builds a constraint for shear lag
        0.5 * bt - 0.5 * tw <= Lpi/50 + sqrt(2)*aw
        """

        a = np.zeros(self.nvars())
        con_type = '<'

        for i in range(len(self.vars)):

            if self.vars[i].target["property"] == "bt" or \
                    self.vars[i].target["property"] == "b":
                a[i] = +0.5
            elif self.vars[i].target["property"] == "tw":
                a[i] = -0.5

        b = self.Lpi / 50 + math.sqrt(2) * mem.cross_section.weld_throat

        con_name = "Top flange shear buckling"

        con = LinearConstraint(a, b, con_type, name=con_name)

        return con

    def BottomFlangeShearLagCon(self, mem):
        """
        Builds a constraint for shear lag
        0.5 * bb - 0.5 * tw <= Lpi/50 + sqrt(2)*aw
        """

        a = np.zeros(self.nvars())
        con_type = '<'

        for i in range(len(self.vars)):

            if self.vars[i].target["property"] == "bb":
                a[i] = +0.5
            elif self.vars[i].target["property"] == "tw":
                a[i] = -0.5

        b = self.Lpi / 50 + math.sqrt(2) * mem.cross_section.weld_throat

        con_name = "Bottom flange shear buckling"

        con = LinearConstraint(a, b, con_type, name=con_name)

        return con

    # def ShearBucklingCon(self, mem):
    #     """
    #     Builds a constraint for shear buckling resistance
    #     hw - (72 * e / ny) * tw <= 0
    #     """
    #
    #     a = np.zeros(self.nvars())
    #     con_type = '<'
    #
    #     e = mem.cross_section.eps
    #
    #     fy =
    #     if fy > 460:
    #         ny = 1
    #     else:
    #         ny = 1.2
    #
    #     for i in range(len(self.vars)):
    #
    #         if self.vars[i].target["property"] == "HW":
    #             a[i] = +1
    #         elif self.vars[i].target["property"] == "tw":
    #             a[i] = -0.5
    #
    #     b = self.Lpi / 50 + math.sqrt(2) * mem.cross_section.weld_throat
    #
    #     con_name = "Shear lag "
    #
    #     con = LinearConstraint(a, b, con_type, name=con_name)
    #
    #     return con

    def create_stability_constraints(self, mem, section_class):
        """
        Creates stability constraint functions
        :return:
        """

        def buckling_y(x):
            self.substitute_variables(x)
            return -mem.ned / mem.NbRd[0] - 1

        def buckling_z(x):
            self.substitute_variables(x)
            return -mem.ned / mem.NbRd[1] - 1

        def com_compression_bending_y(x):
            self.substitute_variables(x)
            return mem.member.check_beamcolumn(section_class=section_class)[0] - 1

        def com_compression_bending_z(x):
            self.substitute_variables(x)
            return mem.member.check_beamcolumn(section_class=section_class)[1] - 1

        def lt_buckling(x):
            self.substitute_variables(x)
            #  med = max(abs(np.min(np.array(mem.steel_member.myed))),
            #  np.max(np.array(mem.steel_member.myed)))
            # print("mem.med = {0:4.2f}".format(mem.med*1e-6),
            #       "mem.MbRd = {0:4.2f}".format(mem.MbRd*1e-6))
            if x[0] < 0:
                print(x)
            return abs(mem.med) / mem.MbRd - 1

        return \
            buckling_y, buckling_z, com_compression_bending_y, \
            com_compression_bending_z, lt_buckling

    def create_section_constraints(self, sect, elem):
        """
        Creates cross-section constraints
        :return:
        """

        def compression(x):
            N = elem.axial_force[0]
            return -N / sect.NRd - 1

        def tension(x):
            N = elem.axial_force[0]
            return N / sect.NRd - 1

        def shear(x):
            V = elem.shear_force[0]
            return abs(V) / sect.VRd - 1

        def bending_moment(x):
            M = elem.bending_moment[0]
            return abs(M) / sect.MRd[0] - 1

        return compression, tension, shear, bending_moment

    def create_constraints(self, section_class):
        """ Create constraints for each member 
            (for the single column the list is a single member)
        """
        for mem in self.structure.members.values():

            
            buckling_y, buckling_z, com_compression_bending_y, \
                com_compression_bending_z, lt_buckling = \
                self.create_stability_constraints(mem, section_class)

            buckling_y_con = NonLinearConstraint(con_fun=buckling_y,
                                                 name="Buckling_y " +
                                                      str(mem.mem_id),
                                                 )
            buckling_y_con.fea_required = True
            
            com_compression_bending_con_y = NonLinearConstraint(
                con_fun=com_compression_bending_y,
                name="Com_compression_bending_y " + str(mem.mem_id),
                )
            com_compression_bending_con_y.fea_required = True
            self.add(com_compression_bending_con_y)

            # self.add(buckling_y_con)
            
            if self.buckling_z:

                buckling_z_con = NonLinearConstraint(con_fun=buckling_z,
                                                     name="Buckling_z " +
                                                          str(mem.mem_id),
                                                     )
                buckling_z_con.fea_required = True

                # self.add(buckling_z_con)

                com_compression_bending_con_z = NonLinearConstraint(
                    con_fun=com_compression_bending_z,
                    name="Com_compression_bending_z " + str(mem.mem_id),
                    )
                com_compression_bending_con_z.fea_required = True
                self.add(com_compression_bending_con_z)
            
            if self.LT_buckling:

                lt_buckling_con = NonLinearConstraint(con_fun=lt_buckling,
                                                      name="LT_buckling " +
                                                           str(mem.mem_id),
                                                      )
                lt_buckling_con.fea_required = True
                self.add(lt_buckling_con)

            """
            for i, elem in enumerate(mem.elements.values()):
                forces = [elem.axial_force[0], elem.shear_force[0],
                          elem.bending_moment[0]]
                compression, tension, shear, bending_moment = \
                    self.create_section_constraints(mem, elem)

                compression_con = NonLinearConstraint(con_fun=compression,
                                                      name="Compression " +
                                                           str(mem.mem_id) +
                                                           str(i), )
                compression_con.fea_required = True

                tension_con = NonLinearConstraint(con_fun=tension,
                                                  name="Tension " +
                                                       str(mem.mem_id) +
                                                       str(i), )
                tension_con.fea_required = True

                shear_con = NonLinearConstraint(con_fun=shear,
                                                name="Shear " + str(mem.mem_id)
                                                     + str(i), )
                shear_con.fea_required = True

                bending_moment_con = NonLinearConstraint(
                    con_fun=bending_moment, name="Bending_moment " +
                                                 str(mem.mem_id) +
                                                 str(i), )
                bending_moment_con.fea_required = True

                self.cons.extend([compression_con, tension_con, shear_con,
                                  bending_moment_con])

                if i == len(mem.elements) - 1:
                    forces = [elem.axial_force[1], elem.shear_force[1],
                              elem.bending_moment[1]]
                    compression, tension, shear, bending_moment = \
                        self.create_section_constraints(mem, elem)

                    compression_con = NonLinearConstraint(con_fun=compression,
                                                          name="Compression " +
                                                               str(mem.mem_id)
                                                               + str(i+1),
                                                          )
                    compression_con.fea_required = True
                    tension_con = NonLinearConstraint(con_fun=tension,
                                                      name="Tension " +
                                                           str(mem.mem_id) +
                                                           str(i+1),
                                                      )
                    tension_con.fea_required = True
                    shear_con = NonLinearConstraint(con_fun=shear,
                                                    name="Shear " +
                                                         str(mem.mem_id) +
                                                         str(i+1),
                                                    )
                    shear_con.fea_required = True

                    bending_moment_con = NonLinearConstraint(
                        con_fun=bending_moment, name="Bending_moment " +
                                                     str(mem.mem_id) +
                                                     str(i+1), )
                    bending_moment_con.fea_required = True

                    self.cons.extend([compression_con, tension_con, shear_con,
                                      bending_moment_con])
            """

            self.add(self.WIColumnWebClassCon(mem))
            self.add(self.WIColumnTopFlangeClassCon(mem))
            self.add(self.TopFlangeShearLagCon(mem))
            if self.symmetry == "mono":
                self.add(self.WIColumnBottomFlangeClassCon(mem))
                self.add(self.BottomFlangeShearLagCon(mem))
                
            self.add(self.WIColumnWebHeightCon(h_min=50))


if __name__ == "__main__":
    from metku.optimization.solvers import *
    from metku.optimization.solvers.bnb import BnB
    from metku.optimization.result_exporter import *

    problem = WIColumn(prob_type='discrete',
                       Lpi=8000,                       
                       top_flange_class=3,
                       bottom_flange_class=3,
                       web_class=3,
                       buckling_z=True,
                       LT_buckling=True)

    # x0 = [300, 8, 200, 10, 200, 10]
    # x0 = [300, 8, 200, 10]
    x0 = [200, 10, 170, 12]

    # x0 = [var.ub for var in problem.vars]

    # # TrustRegionConstr
    # solver = TrustRegionConstr()
    # f_best, x_best, nit = solver.solve(problem,
    #                                    maxiter=200,
    #                                    x0=x0)
    # # print(x_best)
    # problem.num_iters = nit
    # problem(x_best, prec=5)

    # BnB
    lb_solver = TrustRegionConstr()
    solver = BnB(problem, lb_solver)

    # return solver, problem

    # break

    solver.solve(problem, x0=x0, verb=1)

    print(solver.X)
    print(solver.best_x)
    print(solver.best_f)

    # SLP
    # solver = SLP(move_limits=[0.9, 6])
    # solver.solve(problem,
    #              maxiter=50000,
    #              maxtime=30,
    #              x0=x0)
    # problem(solver.X, prec=5)

    # SLSQP
    # solver = slsqp.SLSQP()
    # f_best, x_best = solver.solve(problem,
    #                               maxiter=100,
    #                               x0=x0)
    # problem(solver.best_x, prec=5)

    # MISLP
    # solver = MISLP(move_limits=[0.5, 5])
    # # problem(x0)
    # solver.solve(problem,
    #              maxiter=100,
    #              x0=x0,
    #              min_diff=1e-2,
    #              verb=True)
    # problem(solver.X, prec=5)

    # 2-vaihetekniikalla
    # solver1 = SLP(move_limits=[0.9, 4])
    # solver1 = TrustRegionConstr()
    # solver2 = MISLP(move_limits=[0.85, 1.5])
    # solver = TwoPhase(solver1, solver2, limits=[3, 3])
    # fopt, xopt = solver.solve(problem,
    #                           x0=x0,
    #                           maxiter=50,
    #                           # min_diff=1e-6,
    #                           # verb=True
    #                           )
    # problem(xopt)

    import matplotlib.pyplot as plt

    """
    fvals = problem.fvals
    X = np.arange(len(fvals))
    plt.plot(X, fvals)
    plt.show()
    """

    #ResultExporter(problem, solver).to_csv()
    #ResultExporter(problem, solver).csv_to_excel()

    seconds = time.process_time()
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    print("Process time:", "%d:%02d:%02d" % (h, m, s))

    # problem.structure.members[0].cross_section.draw()

    # G = {}
    # for con in problem.cons:
    #     name = con.name
    #     G[name] = 0

    # print(problem.structure.f.elements[0].bending_moment)
    # print(problem.structure.f.elements[0].axial_force)
    # print(problem.structure.f.loads[1].qval)
    # print("PINTA_ALA=", problem.structure.members[0].A)





# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  3 15:03:13 2022

Various functions for Raami package

@author: kmela
"""

import numpy as np


def divide_line_segment(X1,X2,nseg):
    """ Divides line segment between points X1 and X2 into 'nseg'
        segments.
        
        Returns the points, including X1 and X2
    """
    
    # Determine direction vector from X1 to X2
    #v = (X2-X1)/np.linalg.norm(X2-X1)
    v = X2-X1
    
    T = np.linspace(0,1,nseg+1)
    
    X = [X1 + t*v for t in T]
    
    return X



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Fri Nov 30 11:43:48 2018

@author: huuskoj
"""

import unittest
from ISection import IPE, HEA


# Values from Arcelormittal catalogue, units [mm]
# A, Iy, Wely, Wply, iy, Avz, Iz, Welz, Wplz, iz, ss, It, Iw
IPE200 = [28.5e2, 1943.0e4, 194.0e3, 221.0e3,  8.26, 14.0e2,
          142.0e4, 28.5e3, 44.6e3, 2.24, 36.7, 6.98e4, 13.0e9]
        
ipe = IPE(200)

class TestISection(unittest.TestCase):

    def test_area(self):
        self.assertAlmostEqual(ipe.A, IPE200[0], delta=IPE200[0]*0.01)

    def test_Iy(self):
        self.assertAlmostEqual(ipe.I[0], IPE200[1], delta=IPE200[1]*0.01)

    def test_Wely(self):
        self.assertAlmostEqual(ipe.Wel[0], IPE200[2], delta=IPE200[2]*0.01)

    def test_Wply(self):
        self.assertAlmostEqual(ipe.Wpl[0], IPE200[3], delta=IPE200[3]*0.01)

    """
    def test_iy(self):
        ipe = IPE(200)
        self.assertAlmostEqual(ipe., IPE200[4])
    """

    def test_Avz(self):
        self.assertAlmostEqual(ipe.Ashear, IPE200[5], delta=IPE200[5]*0.01)

    def test_Iz(self):
        self.assertAlmostEqual(ipe.I[1], IPE200[6], delta=IPE200[6]*0.01)

    def test_Welz(self):
        self.assertAlmostEqual(ipe.Wel[1], IPE200[7], delta=IPE200[7]*0.01)

    def test_Wplz(self):
        self.assertAlmostEqual(ipe.Wpl[1], IPE200[8], delta=IPE200[8]*0.01)

    def test_It(self):
        self.assertAlmostEqual(ipe.It, IPE200[-2], delta=IPE200[-2]*0.01)

    def test_Iw(self):
        self.assertAlmostEqual(ipe.Iw, IPE200[-1], delta=IPE200[-1]*0.01)


if __name__ == '__main__':
    unittest.main()



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue May 21 11:00:48 2019

Sequential programming with quadratic objective function.

At each iteration, the problem

min grad f(x_k)'*(x-xk) + 0.5*||x-x_k||**2
such that grad g_j(x_k)'*(x-xk) + g_j(x_k) <= 0

is solved.

@author: kmela
"""


import scipy.optimize as sciop
import numpy as np
import time

import gurobipy as grb


from metku.optimization.solvers.optsolver import OptSolver
from metku.optimization.structopt import *
from metku.optimization.benchmarks import *

    

#from ortools.linear_solver import pywraplp

class SLQP(OptSolver):

    def __init__(self):
        super().__init__()

    def take_action(self):
        """
        Defines action to take
        :return:
        """

        """ linearize problem at X
            A .. Jacobian matrix of the constraints
            B .. RHS of the linearized constraints
            df .. gradient of the objective function
            fx .. f(x_k)
        """
        #print(self.X)
        A, B, df, fx = self.problem.linearize(*self.X)

        # bounds = [x*self.move_limits for x in self.X]
        #
        # res = linprog(df, A, B, bounds=bounds)

        
        qp_fun = lambda x: 0.5*(x-self.X).dot(x-self.X) + df.dot(x-self.X)
        qp_jac = lambda x: (x-self.X) + df
        
        qp_con = {"type":'ineq',"fun":lambda x: A.dot(x) - B,"jac":lambda x: A}

    
        x_bnd = []
        for var in self.problem.vars:
            x_bnd.append((var.lb,var.ub))
            
        print(x_bnd)

        print(qp_fun(self.X))
        print(qp_jac(self.X))
        print(qp_con["fun"](self.X))
        print(qp_con["jac"](self.X))

        qp_res = sciop.minimize(qp_fun,self.X,method='SLSQP',constraints=qp_con,bounds=x_bnd)        
        print(qp_res)

        if qp_res.success == True:
            xnew = qp_res.x
        else:
            print('Something happened with qp problem.')
            xnew = qp_res.x
        
        #print(self.X)
        #return xnew - self.X
        """    
        print(A,B)
        qp_fun = lambda d: 0.5*d.dot(d) + df.dot(d)
        qp_jac = lambda d: d + df
        
        qp_con = {"type":'ineq',"fun":lambda d: A.dot(d) - B}#,"jac":lambda x: A.T}

        d0 = -df
        d_bnd = []
        for i, var in enumerate(self.problem.vars):
            d_bnd.append((var.lb-self.X[i],var.ub-self.X[i]))
            
        #print(x_bnd)

        #print(qp_fun(self.X))
        #print(qp_jac(self.X))
        #print(qp_con["fun"](self.X))
        #print(qp_con["jac"](self.X))

        qp_res = sciop.minimize(qp_fun,d0,method='SLSQP',jac=qp_jac,constraints=qp_con,bounds=d_bnd)    
        print(qp_res)

        if qp_res.success == True:
            dnew = qp_res.x
        else:
            print('Something happened with qp problem.')
            dnew = qp_res.x
            
        return dnew
        """

        
        qp = grb.Model("qp")
        
        # Number of constraints
        n = A.shape[0]
        # Number of variables
        m = A.shape[1]

        x = {}
        #d = {}
        # Variables
        qpvars = []
        qpobj = []
        for i, var in enumerate(self.problem.vars):
            # create variable
            x[i] = qp.addVar(var.lb,var.ub)
            #d[i] = qp.addVar(lb=-200,ub=200)
            # add terms to the objective function
            qpobj.append(0.5*(x[i]-self.X[i])*(x[i]-self.X[i])+df[i]*x[i])
            #qpobj.append(0.5*d[i]*d[i]+df[i]*d[i])
        
        """ Prepare quadratic part of the objective function:
            square terms
        """
        #qpobj = [0.5*va*va for va in qpvars]
        
        
        qp.setObjective(grb.quicksum(qpobj))
        
        # Constraints
        for i in range(n):
            qp.addConstr(grb.quicksum([A[i,j]*x[j] for j in range(m)]) <= B[i])
            #qp.addConstr(grb.quicksum([A[i,j]*d[j] for j in range(m)]) <= B[i])

        
        qp.setParam("LogToConsole",0)
        #qp.update()
        qp.optimize()
        

        X = []
        for v in qp.getVars():
            print('%s %g' % (v.varName, v.x))
            X.append(v.x)
        
        #time.sleep(10)

        X = np.asarray(X)
        
        #self.qp = qp
        
        d = X-self.X
        print(d)
        #return X
        return X-self.X
        
       
        
    def step(self, action):
        
        self.X += action
        for i in range(len(self.X)):
            self.X[i] = np.clip(self.X[i], self.problem.vars[i].lb,
                                self.problem.vars[i].ub)

        self.problem.substitute_variables(self.X)

        return self.X.copy(), 1, False, 'INFO'

if __name__ == '__main__':

    problem = FifteenBarTruss(prob_type='continuous')
    x0 = [var.ub for var in problem.vars]
    solver = SLQP()
    solver.solve(problem, x0=x0, maxiter=100, log=True, verb=True)
    problem(solver.X)
    problem.structure.plot()
    
    
    
    
    
    # #
    # # import matplotlib.pyplot as plt
    # #
    # # plt.plot(solver.fvals)
    # # plt.show()

    # problem = OptimizationProblem("Quadratic Problem")
    # problem.prob_type = 'continuous'
    # problem.obj = lambda x: x[0] ** 2 + x[1] ** 2
    # var1 = Variable("X1", 0, 5)
    # var2 = Variable("X2", 0, 5)
    # problem.add_variables([var1, var2])
    # con1 = NonLinearConstraint(lambda x: x[0] ** 2 / 20 - x[1] + 1)
    # con2 = NonLinearConstraint(lambda x: x[1] ** 2 / 20 - x[0] + 1)
    # problem.add_constraints([con1, con2])
    # solver = SLP()
    # x0 = [0.1, 0.1]
    # solver.solve(problem, x0=x0, maxiter=30)




# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Thu Oct 17 19:53:07 2019

Composite beam sections

@author: kmela
"""

import math
import numpy as np

from copy import deepcopy
import matplotlib.pyplot as plt
import matplotlib.patches as patches


from metku.eurocodes.en1993 import en1993_1_1, en1993_1_5, constants
from metku.eurocodes.en1994 import en1994_1_1
    

class ConcreteFilledTube:
    """ Class for composite columns that are concrete filled tubes
    """
    
    def __init__(self,length,steel_part,concrete,rebar=None,rebar_pos=None):
        """ Constructor
            Input:
                steel_part: RHS or SHS class profile of the steel section
                concrete: Concrete class material
                rebar: 
                
            parameters
                length .. length of the beam
                steel .. steel profile
                concrete .. Concrete class material
                rebar .. Rebar class. Rebar is placed symmetrically by default,
                        using a group of four rebars
                rebar_pos .. coordinates of the top right rebar, assumed to lay
                            in the positive orthant
                nbars .. number of rebars
                ned .. axial force (assumed compression)
                nged .. axial force due to dead loads
                med .. med[0] .. bending moment at the top of the column
                       med[1] .. bending moment at the bottom of the column
                
        """
        
        self.length = length
        self.steel = steel_part
        self.concrete = concrete
        self.rebar = rebar
        self.rebar_pos = [0,0]
        self.nbars = 4
        self.ned = 0.0
        self.nged = 0.0
        self.med = [0.0,0.0]
        
        self.lcr = [1.0,1.0]

        
        if rebar_pos is None:
            """ Set default position """
            self.u = 30
            self.v = 30
        else:    
            self.rebar_pos = rebar_pos
    
    
    @property
    def weight_per_length(self):
        """ Weight per unit length (kN/m) """
        Wsteel = self.steel.weight()*1e3*9.81e-3
        Wconcrete = self.Ac*1e-6*self.concrete.density 
        Wrebar = self.rebar.weight*self.nbars*9.81e-3
        
        #print(Wsteel,Wconcrete,Wrebar)
        return Wsteel + Wconcrete + Wrebar
    
    @property
    def NEd(self):
        """ Design value of axial force """
        return abs(self.ned)
    
    @property
    def M1(self):
        """ Minimum of bending end moments (absolute value) """
        if abs(self.med[0]) <= abs(self.med[1]):
            return self.med[0]
        else:
            return self.med[1]
        
    
    @property
    def M2(self):
        """ Maximum of bending end moments (absolute value) """
        if abs(self.med[0]) <= abs(self.med[1]):
            return self.med[1]
        else:
            return self.med[0]
    
    @property
    def fyd(self):
        """ Design value of yield strength of steel"""
        return self.steel.fy
    
    @property
    def fcd(self):
        """ Design value of yield strength of concrete """
        return self.concrete.fcd(a_cc=1.0)
    
    @property
    def fsd(self):
        """ Design value of yield strength of reinforcement """
        return self.rebar.fyd
    
    @property
    def lcry(self):
        """ Buckling length with respect to major axis """
        return self.length*self.lcr[0]
    
    @property
    def lcrz(self):
        """ Buckling length with respect to minor axis """
        return self.length*self.lcr[1]
    
    @property
    def u(self):
        """ Distance in z-direction (vertical) of the rebar edge from
            inner surface of the steel profile
        """
        return 0.5*self.steel.H - self.steel.T -  (self.rebar_pos[1] + self.rebar.r)
        
    @u.setter
    def u(self,u):
        self.rebar_pos[1] = 0.5*self.steel.H - self.steel.T - u - self.rebar.r
    
    @property
    def v(self):
        """ Distance in y-direction (horizontal) of the rebar edge from
            inner surface of the steel profile
        """
        return 0.5*self.steel.B - self.steel.T - self.rebar_pos[0] + self.rebar.r
        
    @v.setter
    def v(self,v):
        self.rebar_pos[0] = 0.5*self.steel.B - self.steel.T - v - self.rebar.r
    
    @property
    def z_rebar(self):
        """ Vertical position of the rebar (center) from the center of the profile """
        return self.rebar_pos[1]
    
    @property
    def y_rebar(self):
        """ Horizontal position of the rebar (center) from the center of the profile """
        return self.rebar_pos[0]
    
    @property
    def As(self):
        """ Total area of rebar
        """
        return 4*self.rebar.A
    
    @property
    def Aa(self):
        """ Area of steel part """
        return self.steel.A
    
    @property
    def Ia(self):
        """ Second moment of area of steel part with respect to
            the major axis at centroid of steel part
        """
        return self.steel.I[0]
    
    @property
    def Wpa(self):
        
        return self.steel.Wpl[0]
    
    @property
    def Ea(self):
        """ Young's modulus of steel part """
        return self.steel.E
    
    @property
    def EAa(self):
        """ Axial stiffness of steel part """
        return self.Ea*self.Aa
    
    @property
    def EIa(self):
        """ Bending stiffness of steel part 
            with respect to the neutral axis of steel part
        """
        return self.Ea*self.Ia
    
    @property
    def Is(self):
        """ Second moment of area of reinforcement """
        Is0 = self.rebar.I
        Is1 = self.nbars*(Is0 + self.rebar.A*self.z_rebar**2)
        
        return Is1
    
    @property
    def Wps(self):
        """ Bending resistance of reinforcement """
        return self.nbars*self.rebar.A*self.z_rebar
    
    @property
    def EIs(self):
        """ Bending stiffness of reinforcement """
        return self.rebar.E*self.Is
    
    @property
    def Eceff(self):
        """ Effective Young's modulus for concrete 
            EN 1994-1-1, Eq. (6.41)
        """
        phi = self.concrete.creep_coefficient(h0=1e11,t=1e11,t0=28)
        
        print(phi)
        return self.Ecm*1/(1+self.nged/self.ned*phi)
        
    
    @property
    def Ecm(self):
        """ Young's modulus of concrete """
        return self.concrete.Ecm
    
    @property
    def EIc(self):
        """ Bending stiffness of concrete part 
            with respect to the neutral axis of concrete part
        """
        return self.Ecm*self.Ic
    
    @property
    def Ac(self):
        """ Area of concrete part """
        Ac = (self.steel.H-2*self.steel.R)*(self.steel.B-2*self.steel.T) + \
                2*(self.steel.B-2*self.steel.R)*(self.steel.R-self.steel.T) + \
                math.pi*(self.steel.R-self.steel.T)**2 -\
                self.As
        return Ac
        #print(Ac)
        #return (self.steel.H-2*self.steel.T)*(self.steel.B-2*self.steel.T)-self.As
    
    @property
    def Ic(self):
        """ Second moment of area of concrete part with respect to
            major axis at the centroid
        """
        b = self.steel.B
        h = self.steel.H
        t = self.steel.T
        # Inner radius
        r = self.steel.R-t
        
        Icy = (b-2*t)*(h-2*t)**3/12 - r**4/3 - r**2*(h-2*t-r)**2 + \
                +r**4*(9*math.pi**2-64)/36/math.pi + \
                +math.pi*r**2*(0.5*h-t+4*r/3/math.pi)**2 - self.Is
        
        return Icy
    
    
    @property
    def Wpc(self):
        """ Section modulus of concrete """
        b = self.steel.B
        h = self.steel.H
        t = self.steel.T
        # Inner radius
        r = self.steel.R-t
        
        return 0.25*(b-2*t)*(h-2*t)**2 - 2*r**3/3 - r**2*(4-math.pi)*(0.5*h-t-r)-self.Wps

    
    @property
    def EAc(self):
        """ Axial stiffness of concrete """
        return self.Ecm*self.Ac
        

    def EAcom(self):
        """ axial stiffness composite section """
        return self.EAa + self.EAc

    def EIcom(self):
        """ sending stiffness of composite section """        
        return self.Ea*(self.Ia+self.ya_com()**2*self.Aa) + self.Ecm*(self.Ic+self.yc_com()**2*self.Ac)

    
    def EIeff(self,axis="y"):
        """ Effective flexural stiffness for determining Ncr 
            EN 1994-1-1, Eq. (6.40)
        """
        Ke = 0.6
        
        if axis == "y":
            Ic = self.Ic
    
        return self.EIa + self.EIs + Ke*self.Eceff*Ic
    
    
    def EIeffII(self,axis="y"):
        """ Effective flexural stiffness
            EN 1994-1-1, Eq. (6.42)
        """
        Ko = 0.9
        Ke = 0.5
        
        if axis == "y":
            Ic = self.Ic
    
        return Ko*(self.EIa + self.EIs + Ke*self.Eceff*Ic)
    
    def e0(self):
        """ Member imperfection 
            EN 1994-1-1, table 6.5
        """
        
        if self.reinforcement_ratio() <= 0.03:
            r = 300
        else:
            r = 200
            
        return self.length/r
    
    def Ra(self):
        """ Full resultant of steel part """
        
        return self.steel.fy*self.Aa
    
    def Rc(self):
        """ Full resultant of concrete part """
        
        return 0.85*self.slab.material.fcd()*self.Ac
    
    def Rf(self):
        """ Resultant of top flange of the steel beam """
        
        Af = self.steel.tf * self.steel.b        
        
        return self.steel.fy*Af
    
    def Rw(self):
        """ Resultant of web of the steel beam """
        
        Aw = self.steel.hw * self.steel.tw        
        
        return self.steel.fy*Aw
    
    def steel_contribution_ratio(self,verb=False):
        
        delta_a = self.NaRd()/self.NplRd()
        
        if verb:
            print("Steel contribution ratio, delta_a = {0:4.2f}".format(delta_a))
        
        return delta_a
    
    
    def reinforcement_ratio(self,verb=False):
        
        rho_s = self.As/self.Ac
        
        if verb:
            print("Reinforcement ratio, rho_s = {0:4.2f}".format(rho_s))
        
        return rho_s
    
    def NaRd(self,verb=False):
        """ Axial resistance of the steel part"""
        NaRd = self.steel.fy*self.Aa
        
        if verb:
            print("Na,Rd = {0:4.2f} kN".format(NaRd*1e-3))
        return NaRd
    
    def NplRd(self,verb=False):
        """ Axial resistance """
        NplRd = self.steel.fy*self.Aa + self.rebar.fyd*self.As + self.fcd*self.Ac
        
        if verb:
            print("Npl,Rd = {0:4.2f} kN".format(NplRd*1e-3))
        return NplRd
    
    def NplRk(self,verb=False):
        """ Axial resistance with characteristic strength values"""
        NplRk = self.steel.fy*self.Aa + self.rebar.fy*self.As + self.concrete.fck*self.Ac
        
        if verb:
            print("Npl,Rk = {0:4.2f} kN".format(NplRk*1e-3))
            
        return NplRk
    
    def Ncr(self,axis="y"):
        """ Elastic critical force """
        
        if axis == "y":
            Lcr = self.lcry
        else:
            Lcr = self.lcrz
            
        return math.pi**2*self.EIeff(axis)/Lcr**2
    
    def Ncreff(self,axis="y"):
        """ Elastic critical force """
        
        if axis == "y":
            Lcr = self.lcry
        else:
            Lcr = self.lcrz
            
        return math.pi**2*self.EIeffII(axis)/Lcr**2
    
    
    def slenderness(self,axis="y"):
        
        Ncr = self.Ncr(axis)
        
        return math.sqrt(self.NplRk()/Ncr)
    
    def NbRd(self,axis="y",verb=False):
        """ Buckling strength """
        l = self.slenderness(axis)
        if self.reinforcement_ratio() <= 0.03:
            imp_factor = en1993_1_1.buckling_curve["a"]
        else:
            imp_factor = en1993_1_1.buckling_curve["b"]
            
        chi = en1993_1_1.buckling_reduction_factor(l,imp_factor)
    
        NbRd = chi*self.NplRd()
    
        if verb:
            print("**Column buckling: **")
            print("Slenderness: {0:4.3f}".format(l))
            print("Imperfection factor: {0:4.2f}".format(imp_factor))
            print("Buckling reduction factor: {0:4.3f}".format(chi))
            print("Buckling strength: {0:4.2f} kN".format(NbRd*1e-3))
            print("Utilisation: {0:4.2f} ".format(self.NEd/NbRd))
        
        return NbRd 
    
    
    def NpmRd(self):
        
        return self.fcd*self.Ac
    
    def Mpl_max_Rd(self):
        """ Maximum moment resistance """        
        
        return self.fyd*self.Wpa + self.fsd*self.Wps + 0.5*self.fcd*self.Wpc
    
    def MplRd(self,verb=False):
        """ Plastic moment resistance """
        b = self.steel.B
        t = self.steel.T
        fsd = self.fsd
        fcd = self.fcd
        fyd = self.fyd
        
        """ Assume first that the neutral axis is between rebars """
        Asn = 0
        
        hn = (self.NpmRd() - Asn*(2*fsd-fcd))/(2*b*fcd + 4*t*(2*fyd-fcd))
        
        
        """ If neutral axis is below top rebar row, there is no negative
            part for the plastic resistance
        """
        if hn < self.z_rebar - self.rebar.r:
            Wps_n = 0
        else:
            Asn = self.As
            hn = (self.NpmRd() - Asn*(2*fsd-fcd))/(2*b*fcd + 4*t*(2*fyd-fcd))
            Wps_n = self.Wps
            
        Wpc_n = (b-2*t)*hn**2 - Wps_n
        Wpa_n = b*hn**2 - Wpc_n - Wps_n

        MnRd = fyd*Wpa_n + fsd*Wps_n + 0.5*fcd*Wpc_n

        MplRd = self.Mpl_max_Rd() - MnRd 

        if verb:
            print("Plastic moment resistance:")
            print("hn = {0:4.2f} mm (z_rebar = {1:4.2f} mm)".format(hn,self.z_rebar))
            print("Wps_n = {0:4.2f} mm3".format(Wps_n))
            print("Wpc_n = {0:4.2f} mm3".format(Wpc_n))
            print("Wpa_n = {0:4.2f} mm3".format(Wpa_n))
            print("MplRd = {0:4.2f} kNm".format(MplRd*1e-6))

        return MplRd 
    
    def alpha_m(self):
        
        if self.steel.fy <= 355:
            alpha_m = 0.9
        else:
            alpha_m = 0.8
            
        return alpha_m
    
    
    
    def VRd(self):
        """ Shear resistance
            This equals shear resistance of the steel part
        """
        
        return self.steel.shear_force_resistance()
    
    def design(self,verb=False):
        """ Carry out design according to EN 1994-1-1, simplified method """
        
        """ Check validity """
        side_ratio = self.steel.H/self.steel.B
        
        A_ratio = self.reinforcement_ratio()
        
        if side_ratio < 0.2 or side_ratio > 5.0:
            """ steel part dimensions out of scope """
            print("Warning: the ratio h/b = {0:4.2f} is out of scope.".format(side_ratio))
        
        if A_ratio > 0.06:
            """ too much rebar """
            print("Warning: too much rebar - {0:4.2f} > 0.06".format(A_ratio))
        
        self.steel_contribution_ratio(verb)
        
        slend = self.slenderness()
        
        if verb:
            print("Slenderness = {0:4.3f}".format(slend))
            
        if slend > 2.0:
            """ too slender member """
            print("Warning: slenderness too large - {0:4.2f} > 2.0".format(A_ratio))
        
        
        if self.M2 == 0:
            """ Only axial force """
            self.NbRd(verb=True)
        
        """ Design for combined bending moment and axial force """
        e0 = self.e0()
        Med0 = self.NEd*e0
        
        if verb:
            print("Initial imperfection e0 = {0:4.2f} mm".format(e0))
            print("Bending moment due to Initial imperfection MEd0 = {0:4.2f} kNm".format(Med0*1e-6))
            
        """ Check for need to include second-order effects """
        Ncr_eff = self.Ncreff()
        
        if verb:
            print("Check for second-order effects:")
            print("Ncr,eff = {0:4.2f} kN".format(Ncr_eff*1e-3))
            print("10*NEd = {0:4.2f} kN".format(10*self.NEd*1e-3))
        
        if Ncr_eff < 10*self.NEd:
            if verb:
                print("Include second-order effects.")
                
            if self.M2 > 0:
                bm = max(0.66+0.44*self.M1/self.M2,0.44)
            else:
                bm = 0
                
            MEdI = bm*self.M2 + Med0
            km = max(1/(1-self.NEd/Ncr_eff),1.0)
            
            MEd = max(km*MEdI,self.M2)
            
            if verb:
                print("bm = {0:4.3f}".format(bm))
                print("MEd,I = {0:4.2f} kNm".format(MEdI*1e-6))
                print("km = {0:4.3f}".format(km))
                print("MEd = {0:4.2f} kNm".format(MEd*1e-6))
                            
        else:
            MEd = Med0 + self.M2
            if verb:
                print("No need to include second-order effects.")
                
        
        """ Do M+N -interaction """
        MplRd = self.MplRd(True)
        NplRd = self.NplRd()
        chiC = self.NpmRd()/NplRd
        chiD = self.NEd/NplRd
        muD = min((chiD-1)/(chiC-1),1.0)
        
        Mpl_N_Rd = muD*MplRd
        
        UMN = MEd/Mpl_N_Rd/self.alpha_m()
        
        if verb:
            print("chi_C = Npm_Rd/Npl_Rd = {0:4.3f}".format(chiC))
            print("chi_d = NEd/Npl_Rd = {0:4.3f}".format(chiD))
            print("mu_d = (chi_d-1)/(chi_C-1) = {0:4.3f}".format(muD))
            print("MplRd = {0:4.3f} kNm".format(MplRd*1e-6))
            print("Mpl_N_Rd = mu_d*Mpl_Rd = {0:4.3f} kNm".format(Mpl_N_Rd*1e-6))
            print("MEd/Mpl_N_Rd = {0:4.3f}".format(MEd/Mpl_N_Rd))
            print("Utilization ratio: {0:4.3f}".format(UMN))
            
        
        
        
    
    def draw(self):
        """ Draw the profile             
        """
        
        fig, ax = plt.subplots(1)

        self.steel.draw(axes=ax)        
                    
        # Draw rebar
        r = self.rebar.r        
        
        xy = np.array((self.rebar_pos))
        
        dxy = np.array([[1,1],[-1,1],[-1,-1],[1,-1]])
        
        for xyloc in dxy:
            ax.add_patch(patches.Circle(tuple(xyloc*xy),r))
        
        
        ax.set_aspect('equal')

if __name__ == '__main__':
    
    from metku.sections.steel.RHS import SHS
    from metku.eurocodes.en1992 import en1992_1_1, constants
    
    steel = SHS(300,8)
    con = constants.Concrete("C40/50")
    rebar = en1992_1_1.Rebar(20,"B500B")
    ned = 2.4e6
    nged = 1.7e6
    med = [0,0]
    
    p = ConcreteFilledTube(6500,steel,con,rebar)
    p.u = 30
    p.ned = 2.118e6
    p.nged = 767.696e3
    p.design(True)
    #p.NplRd(True)
    #p.NaRd(True)
    #p.NplRk(True)
    p.steel_contribution_ratio(True)
    #p.draw()
    


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 23 17:12:02 2020

Steepest descent method

For unconstrained optimization

@author: kmela
"""

import numpy as np

from scipy.optimize import minimize_scalar
import  matplotlib.pyplot as plt

import metku.optimization.structopt as sopt
from metku.optimization.solvers.optsolver import OptSolver

GRAD_TOL = 1e-8
ABS_F_TOL = 1e-8
REL_F_TOL = 1e-4

def line_search_obj(fun,xk,dk):
    """ Objectve function for line search """
    
    def line_fun(a):
        #print(xk + a*dk)
        return fun(xk + a*dk)
    
    return line_fun

class SteepestDescent(OptSolver):
    """ Steepest descent method """

    def __init__(self, fun, grad):
        """ Input:
            fun .. function to be minimized.
            grad .. gradient of the function. If None is given, gradient is
                    evaluated numerically
                    
            TODO:
                implement numerical gradient evaluation!
        """
        super().__init__()
        self.result = None
        self.obj = fun
        self.grad = grad
        
    def solve(self, x0, maxiter=200):
        """ Perform steepest descent iteration """
        
        k = 0
        x = [x0]
        d = []
        
        self.xvals.append(x0)
        self.fvals.append(self.obj(x0))
        
        f_stop = 0
        
        while k < maxiter:
            
            df = self.grad(x[k])
            
            ndf = np.linalg.norm(df,2)
            
            """ Check gradient stopping criterion """
            if ndf < GRAD_TOL:
                break
            else:
                d.append(-df/ndf)
            
            """ Perform line search """
            
            """
            ls_fun = line_search_obj(self.obj,x[k],d[k])
            a = np.linspace(0,5)
            y = np.zeros(a.shape)
            for i in range(a.shape[0]):
                y[i] = ls_fun(a[i])
            
            plt.plot(a,y)
            """
            
            ls_res = minimize_scalar(line_search_obj(self.obj,x[k],d[k]),
                                     method='brent')
            
            #print(d[k])
            #print("Line search result:", ls_res.x)
            #print(ls_res.success)
            if ls_res.success:
                x.append(x[k] + ls_res.x*d[k])
                self.xvals.append(x[-1])
                self.fvals.append(self.obj(x[-1]))
                
            
            #print(x)
            if abs(self.fvals[-1]-self.fvals[-2]) <= ABS_F_TOL + REL_F_TOL*abs(self.fvals[-2]):
                f_stop += 1
                
            if f_stop == 2:
                break
            
            k += 1

def steep_ex_fun(x):
    
    return (x[0]-4)**2+ (4*x[1]-2)**2

def steep_ex_grad(x):
    
    return np.array([2*(x[0]-4),8*(4*x[1]-2)])


if __name__ == "__main__":
    
    steep_sol = SteepestDescent(steep_ex_fun,steep_ex_grad)
    
    steep_sol.solve(np.array([10.0,10.0]),maxiter=15)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
import numpy as np

class LineLoad(Load):
    """ Class for 2D line loads (forces and moments)
    """

    def __init__(self, sid, eid, xloc, qval, direction, coords=1):
        """ Input:
            sid -- load id
            eid -- element subjected to the load
            xloc -- starting and ending locations of the load (local coordinates)
                    xloc[0] .. staring coordinate 0 <= xloc[0] <= 1
                    xloc[1] .. ending coordinate 0 <= xloc[0] < xloc[1] <= 1
            qval -- value of the load at the coordinates xloc
            direction -- 0 .. x-axis
                         1 .. y-axis
            coords .. 0 .. local
                      1 .. global
        """
        Load.__init__(self, sid)

        self.elem = eid
        self.xloc = xloc
        self.qval = qval
        self.dir = direction
        self.coords = coords

    def load_and_dofs(self):
        """ Returns the loads to be inserted to the global load vector
            and the corresponding global degrees of freedom
        """

        """ Get load vector """
        if self.dir == "x":
            q = np.array([self.qval[0], 0, 0])
        else:
            q = np.array([0, self.qval[0], 0])

        """ Get equivalent nodal loads """
        F = self.elem.equivalent_nodal_loads(q)

        """ Number of degrees of freedom """
        dofs = np.array(self.elem.global_dofs())

        """ Find free degrees of freedom (those with numbering
                                          greater than -1)
        """
        nzero_dofs = dofs >= 0

        """ Get free degrees of freedom and the corresponding part
            of the load vector
        """
        d = dofs[nzero_dofs]
        Fnz = F[nzero_dofs]

        return Fnz, d



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
import numpy as np


from metku.optimization.solvers import OptSolver
from metku.optimization.structopt import DiscreteVariable, Variable

"""
phase 1:
- continuous / discrete variables

phase 2:
- discrete / index variables

TwoPhase(solver1, solver2, limits=(1, 3), var_type1='', var_type2="", maxiter1=100, maxiter2=100)

fopt, xopt = solver.solve(problem, x0=x0, maxiter=100, verb=True, log=True)

"""

class TwoPhase:


    def __init__(self, solver1, solver2, limits=(1, 3),
                 var_type1="continuous", var_type2="discrete",
                 maxiter1=None, maxiter2=None):
        super().__init__()

        if not isinstance(solver1, OptSolver)\
                and isinstance(solver2, OptSolver):
            raise ValueError("Solvers must be 'OptSolver' class solvers!")

        self.solver1 = solver1
        self.solver2 = solver2
        self.limits = np.asarray(limits)
        self.var_type1 = var_type1
        self.var_type2 = var_type2
        self.maxiter1 = maxiter1
        self.maxiter2 = maxiter2

        self.problem = None

    def discretize(self):
        """
        Discretizes relaxed problem
        :return:
        """
        # Continuous solution
        best_x = self.solver1.X
        # Initialize problem as discrete problem
        self.problem.__init__(prob_type=self.var_type2)
        for i, var in enumerate(self.problem.vars):
            # if isinstance(var, DiscreteVariable):
            #     pass
            if type(var) is Variable:
                pass
            else:
                values = np.asarray(var.values)
                idx = np.argmin((values - best_x[i])**2)
                var.substitute(var.values[idx])
                ll, ul = self.limits
                lb = max(0, idx - ll)
                ub = min(len(values), idx + ul)
                var.__init__(var.name, var.values[lb:ub], var.target)
            
    def solve(self, problem, **kwargs):
        """
        Solves given problem
        :param problem:
        :param kwargs:
        :return:
        """
        self.problem = problem
        if problem.prob_type != self.var_type1:
            problem.__init__(prob_type=self.var_type1)
        print(f"Solving first phase using {type(self.solver1).__name__}")
        if self.maxiter1 is not None:
            kwargs['maxiter'] = self.maxiter1
        self.solver1.solve(problem, **kwargs)
        self.discretize()
        print(f"Discretizing first phase solution using {type(self.solver2).__name__}")
        x0 = [var.lb for var in self.problem.vars]
        kwargs['x0'] = x0
        if self.maxiter2 is not None:
            kwargs['maxiter'] = self.maxiter2
        xopt, fopt = self.solver2.solve(problem, **kwargs)
            
        return xopt, fopt


if __name__ == '__main__':
    from metku.optimization.solvers import *
    from metku.optimization.benchmarks import *
    from metku.optimization.problems import WIColumn
    import deap

    #solver1 = SLP(move_limits=[0.9, 1.5])
    solver1 = MISLP(move_limits=[0.85, 1.5])
    solver2 = GA(pop_size=100, mutate=deap.tools.mutShuffleIndexes,
                 mutation_kwargs={'indpb':0.15})

    #solver = TwoPhase(solver1, solver2, limits=[3, 3], var_type2='index', maxiter1=40, maxiter2=50)
    problem = WIColumn(prob_type='discrete')
    x0 = [var.ub for var in problem.vars]
    problem(x0)
    #fopt, xopt = solver1.solve(problem, x0=x0, maxiter=50, verb=True)
    #problem(xopt)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 23:12:31 2018

Rectangular hollow sections

@author: kmela
"""

import math
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines


from metku.eurocodes.en1993 import en1993_1_1
from metku.sections.steel.steel_section import SteelSection


class RHS(SteelSection):
    """ Rectangular hollow sections """

    def __init__(self, H, B, T, fy=355):
        """ Rectangular hollow sections
            
            H -- height
            B -- width
            T -- wall thickness
            fy -- yield strength
        """
        R = RHS_outer_radius(T)
        A = RHS_area(H, B, T, R)
        I = [0.0, 0.0]
        I[0] = RHS_second_moment(H, B, T, R)
        I[1] = RHS_second_moment(B, H, T, R)
        Ashear = [RHS_shear_area(A,B,H),RHS_shear_area(A, H, B)]
        
        Au = RHS_paint_area(H, B, R)
        Wel = [0.0, 0.0]
        Wel[0] = RHS_elastic_modulus(H, I[0])
        Wel[1] = RHS_elastic_modulus(B, I[1])
        Wpl = [0.0, 0.0]
        Wpl[0] = RHS_plastic_modulus(H, B, R, T)
        Wpl[1] = RHS_plastic_modulus(B, H, R, T)

        super().__init__(fy,
                         A,
                         I,
                         Au,
                         Wpl,
                         Wel,
                         Ashear)

        self.H = H
        self.B = B
        self.T = T
        self.Iw = 0.0
        self.It = self.torsional_constant()
        self.imp_factor = [en1993_1_1.buckling_curve["c"],
                           en1993_1_1.buckling_curve["c"]]
        self.imp_factor_LT_gen = en1993_1_1.buckling_curve["c"]  # TODO
        self.imp_factor_LT = en1993_1_1.buckling_curve["c"]  # TODO
        
    @property
    def h(self):
        return self.H
    
    @property
    def b(self):
        return self.B

    @property
    def t(self):
        return self.T

    @property
    def r(self):
        return self.R
    
    def __repr__(self):
        return f"{type(self).__name__} {self.H:.0f}X{self.B:.0f}X{self.T:.1f}"
    
    def robot(self):
        
        return f"RRHS {self.H:.0f}x{self.B:.0f}x{self.T:.1f}"

    def info(self,latex=False):
        """ Prints a bunch of section properties """
        
        if latex:
            print("*** " + self.__repr__() + " ***")
            print(" $h = {0:.0f}{1:s}$".format(self.H,'\\,\\unit{mm}'))
            print(" $b = {0:.0f}{1:s}$".format(self.B,'\\,\\unit{mm}'))
            print(" $t = {0:.1f}{1:s}$".format(self.t,'\\,\\unit{mm}'))            
            print(" $r = {0:.1f}{1:s}$".format(self.r,'\\,\\unit{mm}'))
            print(" $A = {0:.2f}{1:s}$".format(self.A*1e-2,'\\ee{2}\\,\\squnit{mm}'))
            print(" $I_y = {0:.2f}{1:s}$".format(self.I[0]*1e-4,'\\ee{4}\\,\\quunit{mm}'))
            print(" $I_z = {0:.2f}{1:s}$".format(self.I[1]*1e-4,'\\ee{4}\\,\\quunit{mm}'))
            print(" $W_el,y = {0:.2f}{1:s}$".format(self.Wel[0]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $W_el,z = {0:.2f}{1:s}$".format(self.Wel[1]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $W_pl,y = {0:.2f}{1:s}$".format(self.Wpl[0]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $W_pl,z = {0:.2f}{1:s}$".format(self.Wpl[1]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))        
            print(" $Ashear = {0:.2f}{1:s}$".format(self.Ashear*1e-2,'\\ee{2}\\,\\squnit{mm}'))
        else:
            print("*** " + self.__repr__() + " ***")
            print(" h = {0:.0f} mm".format(self.h))
            print(" b = {0:.0f} mm".format(self.b))            
            print(" t = {0:.1f} mm".format(self.t))
            print(" r = {0:.1f} mm".format(self.r))
            print(" A = {0:.2f} (100 mm2)".format(self.A*1e-2))
            print(" Iy = {0:.2f} (10^4 mm4)".format(self.I[0]*1e-4))
            print(" Iz = {0:.2f} (10^4 mm4)".format(self.I[1]*1e-4))
            print(" Wel,y = {0:.2f} (10^3 mm4)".format(self.Wel[0]*1e-3))
            print(" Wel,z = {0:.2f} (10^3 mm4)".format(self.Wel[1]*1e-3))
            print(" Wpl,y = {0:.2f} (10^3 mm4)".format(self.Wpl[0]*1e-3))
            print(" Wpl,z = {0:.2f} (10^3 mm4)".format(self.Wpl[1]*1e-3))
            print(" Ashear = {0:.2f} (100 mm2)".format(self.Ashear*1e-2))

    def update_properties(self):
        """
        Updates cross-sectional properties
        """

        try:
            # Using values from catalog, all catalog values
            # have maximum of 1 decimal
            # if (self.H * 10).is_integer() and (self.B * 10).is_integer() and (self.T * 10).is_integer():
            self.R = RHS_outer_radius(self.T)
            # else:
            # self.R = 2 * self.T
            self.A = RHS_area(self.H, self.B, self.T, self.R)
            I = [0.0, 0.0]
            I[0] = RHS_second_moment(self.H, self.B, self.T, self.R)
            I[1] = RHS_second_moment(self.B, self.H, self.T, self.R)
            self.I = np.asarray(I)
            self.Ashear = [RHS_shear_area(self.A, self.B, self.B),RHS_shear_area(self.A, self.H, self.B)]
            self.Au = RHS_paint_area(self.H, self.B, self.R)
            Wel = [0.0, 0.0]
            Wel[0] = RHS_elastic_modulus(self.H, I[0])
            Wel[1] = RHS_elastic_modulus(self.B, I[1])
            self.Wel = np.asarray(Wel)
            Wpl = [0.0, 0.0]
            Wpl[0] = RHS_plastic_modulus(self.H, self.B, self.R, self.T)
            Wpl[1] = RHS_plastic_modulus(self.B, self.H, self.R, self.T)
            self.Wpl = np.asarray(Wpl)
            self.It = self.torsional_constant()

        except AttributeError:
            pass

    def __setattr__(self, key, value):

        dim_vars = ['H', 'B', 'T']
        super().__setattr__(key, value)
        # Whenever dimensions change, update cross-sectional properties
        if key in dim_vars:
            self.update_properties()

    @property
    def c_web(self):
        return self.H - 2 * self.R

    @property
    def c_flange(self):
        return self.B - 2 * self.R

    def torsional_constant(self):
        """
        Calculated according to EN 10219-2
        
        """

        # Rc .. corner radius at center line
        Rc = self.R - 0.5 * self.T
        h = 2*((self.B-self.T) + (self.H-self.T)) - 2*Rc*(4-math.pi)
        Ah = (self.B-self.T) * (self.H-self.T) - Rc**2*(4-math.pi)
        K = 2*Ah*self.T/h
        It = (self.T**3*h/3 + 2*K*Ah)

        return It


    def flange_class(self, verb=False):
        """ Determine class of compressed flange """
        cf = self.B - 2 * self.R
        rf = cf / self.T
        cFlange = en1993_1_1.internal_part_in_compression(rf, self.eps)
        
        if verb:
            print("Flange classification (internal element):")
            print("cf = {0:4.2f}, tf = {1:4.2f}".format(cf,self.T))
            print("cf/tf = {0:4.2f}".format(rf))
            print("Flange class = {0}".format(cFlange))

        return cFlange

    def web_class_comp(self, verb=False):
        """ Determine class of compressed web """
        cw = self.H - 2 * self.R
        rw = cw / self.T
        cWeb = en1993_1_1.internal_part_in_compression(rw, self.eps)

        if verb:
            print("Web classification (internal part in compression):")
            print("cw = {0:4.2f}, tw = {1:4.2f}".format(cw,self.T))
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_bend(self, verb=False):
        """ Determine class of web in bending """
        cw = self.H - 2 * self.R
        rw = cw / self.T
        cWeb = en1993_1_1.internal_part_in_bending(rw, self.eps)
        
        if verb:
            print("Web classification (internal part in bending):")
            print(f"cw = {cw:4.2f}, tw = {self.T:4.2f}")
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb    

    def web_class_comp_bend(self, Ned, verb=False):

        return 2

    def moment_axial_force_interact(self, UN, MRd):
        """ Interaction rule for section resistance for combined
            axial force and bending
            
            input: UN .. NEd/NRd
                  MRd .. moment resistance
        """
        aw = min((self.A - 2 * self.B * self.T) / self.A, 0.5)

        if UN > 1.0:
            MNRd = 0.0
        else:
            MNRd = MRd * min((1 - UN) / (1 - 0.5 * aw), 1)

        return MNRd
    
    def draw(self,axes=None,theta=0):
        """ Draw the profile 
            input:
                axes .. matplotlib.axes.Axes object. If this is given,
                        the profile will be drawn to that Axes
                theta .. rotation (optional)
        """
        
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        if theta != 0:
            phi = np.radians(theta)
            c, s = np.cos(phi), np.sin(phi)
            R = np.array(((c,-s), (s, c)))
        
        lw = 1.5
        col = 'b'
        
        h = self.H
        b = self.B
        t = self.T
        r = self.R
        
        # Draw webs
        y_web = [-0.5*h+r,0.5*h-r]                            
        ax.vlines([-0.5*b,-0.5*b+t,0.5*b-t,0.5*b],y_web[0],y_web[1],colors='b',linewidth=lw)
        
        # Draw flanges
        x_flange = [-0.5*b+r,0.5*b-r]
        y_top = [0.5*h,0.5*h-t]
        y_bottom = [-0.5*h,-0.5*h+t]
        y_flanges = y_top + y_bottom
        
        ax.hlines(y_flanges,x_flange[0],x_flange[1],colors='b',linewidth=lw)
        
        # Draw corners
        # Upper left corner
        x_up_left = (-0.5*b+r,0.5*h-r)
        
        up_left_corner_in = patches.Arc(x_up_left,2*(r-t),2*(r-t),0,90,180,color=col,linewidth=lw)
        up_left_corner_out = patches.Arc(x_up_left,2*r,2*r,0,90,180,color=col,linewidth=lw)
        
        ax.add_patch(up_left_corner_in)
        ax.add_patch(up_left_corner_out)
        
        x_bot_left = (-0.5*b+r,-0.5*h+r)
        
        bot_left_corner_in = patches.Arc(x_bot_left,2*(r-t),2*(r-t),0,180,270,color=col,linewidth=lw)
        bot_left_corner_out = patches.Arc(x_bot_left,2*r,2*r,0,180,270,color=col,linewidth=lw)
        
        ax.add_patch(bot_left_corner_in)
        ax.add_patch(bot_left_corner_out)
                            
        # Upper left corner
        x_up_right = (0.5*b-r,0.5*h-r)
        
        up_right_corner_in = patches.Arc(x_up_right,2*(r-t),2*(r-t),0,0,90,color=col,linewidth=lw)
        up_right_corner_out = patches.Arc(x_up_right,2*r,2*r,0,0,90,color=col,linewidth=lw)
        
        ax.add_patch(up_right_corner_in)
        ax.add_patch(up_right_corner_out)
        
        
        x_bot_right = (0.5*b-r,-0.5*h+r)
        
        bot_right_corner_in = patches.Arc(x_bot_right,2*(r-t),2*(r-t),0,270,360,color=col,linewidth=lw)
        bot_right_corner_out = patches.Arc(x_bot_right,2*r,2*r,0,270,360,color=col,linewidth=lw)
        
        ax.add_patch(bot_right_corner_in)
        ax.add_patch(bot_right_corner_out)
        
        
        a = 0.1
        
        ax.set_xlim(-(0.5+a)*b, (0.5+a)*b)
        ax.set_ylim(-(0.5+a)*h, (0.5+a)*h)
        
        ax.set_aspect('equal')
    
    def abaqus(self,filename,matname='Steel',setname='Set-1',open_type='w'):
        """ Writes section data for exporting to abaqus """
        unit_conv = 1e-3
        
        # Calculate coordinates for abaqus cross-section
        W = self.B*unit_conv
        H = self.H*unit_conv
        T = self.T*unit_conv
        NomT = self.T*unit_conv
        R = self.R*unit_conv
        
        RMid=R-T/2
            
        angle=math.radians(22.5)
        
        X=[]
        Y=[]
        
        X.append(0.0)
        Y.append((H-T)/2)
        
        X.append(W/2-R-2*T)
        Y.append((H-T)/2)  
        
        X.append(W/2-R)
        Y.append((H-T)/2)
        
        X.append(W/2-R+RMid*math.sin(angle))
        Y.append((H-T)/2-(RMid-RMid*math.cos(angle)))
        
        X.append(W/2-R+RMid*math.sin(2*angle))
        Y.append((H-T)/2-(RMid-RMid*math.cos(2*angle))) 
        
        X.append(W/2-R+RMid*math.sin(3*angle))
        Y.append((H-T)/2-(RMid-RMid*math.cos(3*angle))) 
        
        X.append((W-T)/2)
        Y.append(H/2-R) 
        
        X.append((W-T)/2)
        Y.append(H/2-R-2*T) 
        
        X.append((W-T)/2)
        Y.append(0.0)
        

        X += X[7::-1]
        Y += [-y for y in Y[7::-1]]
        
        X += [-x for x in X[1:9]]
        Y += [-y for y in Y[1:9]]
        
        X += [-x for x in X[7:0:-1]]
        Y += Y[7:0:-1]
        
        X.append(X[0])
        Y.append(Y[0])
        
        # Write coordinates to file
        MatName = matname + "_Mat"
        nx = len(X)
        with open(filename,open_type) as file:
            #file.write(f"** Member {mem_name}\n")
            file.write("** Cross-section dimensions \n")
            file.write(f"** WIDTH = {W}\n")
            file.write(f"** HEIGHT = {H}\n")
            file.write(f"** THICKNESS = {T}\n")
            file.write(f"** CORNER RADIUS = {R}\n**\n")
            file.write(f"*Beam Section, elset={setname}, material={MatName}, temperature=GRADIENTS, section=ARBITRARY\n")
            file.write(f'{nx-1}, {X[0]:8.6f},{Y[0]:8.6f},{X[1]:8.6f},{Y[1]:8.6f},{T:8.6f}\n')                 
                   
            for x, y in zip(X[2:],Y[2:]):
                file.write(f"{x:8.6f},{y:8.6f},{T:8.6f}\n")
        
            file.write("0.,1.,0.\n")
        

class SHS(RHS):
    """ Square hollow sections
        Subclass of RHS
    """

    def __init__(self, H, T, fy=355):
        RHS.__init__(self, H, H, T, fy)

    def __repr__(self):
        return f"{type(self).__name__} {self.H:.0f}X{self.B:.0f}X{self.T:.1f}"


# Functions for computing RHS section properties
def RHS_area(H, B, T, R):
    """ cross-sectional area """
    A = 2 * (H - 2 * R) * T + 2 * (B - 2 * R) * T + math.pi * (
                R ** 2 - (R - T) ** 2)
    return A


def RHS_paint_area(H, B, R):
    """ circumference """
    Au = 2 * (H + B) - 8 * R + 2 * math.pi * R
    return Au


def RHS_shear_area(A, H, B):
    Ashear = A * H / (H + B)
    return Ashear


def RHS_second_moment(H, B, T, R):
    """ Second moment of area """
    I1 = 2 * (1 / 12 * T * (H - 2 * R) ** 3 + 1 / 12 * (B - 2 * R) * T ** 3 + (
                0.5 * (H - T)) ** 2 * (B - 2 * R) * T)
    IR = 4 * ((math.pi / 16 - 4 / 9 / math.pi) * R ** 4 + (
                0.5 * H - R + 4 * R / 3 / math.pi) ** 2 * math.pi * R ** 2 / 4)
    IRneg = 4 * ((math.pi / 16 - 4 / 9 / math.pi) * (R - T) ** 4 + (
                0.5 * H - R + 4 * (R - T) / 3 / math.pi) ** 2 * math.pi * (
                             R - T) ** 2 / 4)
    I1 = I1 + IR - IRneg

    return I1


def RHS_elastic_modulus(H, I1):
    Wel = I1 / (0.5 * H)

    return Wel


def RHS_plastic_modulus(H, B, R, T):
    Wp = 2 * ((0.5 * H - R) * T * (0.5 * H - R)) + (B - 2 * R) * T * (H - T)
    WR = 2 * math.pi / 4 * R ** 2 * 2 * (0.5 * H - R + 4 * R / 3 / math.pi)
    Wneg = 2 * math.pi / 4 * (R - T) ** 2 * 2 * (
                0.5 * H - R + 4 * (R - T) / 3 / math.pi)
    Wp = Wp + WR - Wneg

    return Wp


def RHS_outer_radius(T):
    if T <= 6.0:
        R = 2 * T
    elif T <= 10.0:
        R = 2.5 * T
    else:
        R = 3 * T
    return R

if __name__ == "__main__":
    
    p = RHS(250,150,10)
    #p.draw()



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Wed Mar 14 22:50:51 2018

Constants for EN 1992

@author: kmela
"""

# Constants: Young's modulus, Shear modulus, density
# Units: N, mm, i.e. MPa =N/mm^2

# Young's modulus Mpa
E = 210.0e3

# Shear modulus Mpa
G = 81.0e3

# Poisson's ratio MPa
NU = 0.3

# Density kg/mm^3
density = 7850.0e-9

# Partial safety factors
gammaC = 1.5    # Concrete
gammaS = 1.15   # Steel
gammaV = 1.25   # shear stud

# EN 1992-1-1 3.1.6
alpha_cc = 0.85 


""" Material data
fck .. lieriölujuus
fck_cube .. kuutiolujuus
fcm
Ecm
"""
concrete = {"C12/15": {"fck": 12.0, "fck_cube": 15.0, "fcm": 20, "Ecm": 27000.0},
            "C16/20": {"fck": 16.0, "fck_cube": 20.0, "fcm": 24, "Ecm": 29000.0},
            "C20/25": {"fck": 20.0, "fck_cube": 25.0, "fcm": 28, "Ecm": 30000.0},
            "C25/30": {"fck": 25.0, "fck_cube": 30.0, "fcm": 33, "Ecm": 31000.0},
            "C30/37": {"fck": 30.0, "fck_cube": 37.0, "fcm": 38, "Ecm": 33000.0},
            "C35/45": {"fck": 35.0, "fck_cube": 45.0, "fcm": 43, "Ecm": 34000.0},
            "C40/50": {"fck": 40.0, "fck_cube": 50.0, "fcm": 48, "Ecm": 35000.0},
            }

class Concrete:
    
    def __init__(self,material="C25/30"):
        """ Constructor for Concrete material class """

        self.name = material
        self.fck = concrete[material]["fck"]
        self.fck_cube = concrete[material]["fck_cube"]
        self.fcm = concrete[material]["fcm"]
        self.Ecm = concrete[material]["Ecm"]
        
    def fcd(self,a_cc=alpha_cc,gC=gammaC):
        """ Design value for compressive strength """
        return a_cc*self.fck/gC


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-

from metku.optimization.structopt import *
from metku.frame2d.frame2d import *
from metku.frame2d.simple_truss import *


import numpy as np

TEN_BAR_AREAS_in2 = [1.62, 1.80, 1.99, 2.13, 2.38, 2.62, 2.63, 2.88, 2.93,
                    3.09, 3.13, 3.38, 3.47, 3.55, 3.63, 3.84, 3.87, 3.88,
                    4.18, 4.22, 4.49, 4.59, 4.80, 4.97, 5.12, 5.74, 7.22,
                    7.97, 11.50, 13.50, 13.90, 14.20, 15.50, 16.00, 16.90,
                    18.80, 19.90, 22.00, 22.90, 26.50, 30.00, 33.50]


TEN_BAR_AREAS_mm2 = [1044.9, 1161.0, 1283.55, 1373.85, 1535.1, 1689.9,
                     1696.35, 1857.6, 1889.85, 1993.05, 2018.85, 2180.1,
                     2238.15, 2289.75, 2341.35, 2476.8, 2496.15, 2502.6,
                     2696.1, 2721.9, 2896.05, 2960.55, 3096.0, 3205.65,
                     3302.4, 3702.3, 4656.9, 5140.65, 7417.5, 8707.5, 8965.5,
                     9159.0, 9997.5, 10320.0, 10900.5, 12126.0, 12835.5,
                     14190.0, 14770.5, 17092.5, 19350.0, 21607.5]
# Problem parameters
L = 9144  # mm
F = 444890  # N
E = 68950  # MPa
rho = 2768e-9  # kg/mm3

# Constraint limits
sigma_max = 172.37  # MPa
delta_max = 50.8  # mm

class SimpleTenBarTruss(OptimizationProblem):

    # Problem parameters
    L = 9144    # mm
    F = 444890  # N
    E = 68950   # MPa
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 172.37  # MPa
    delta_max = 50.8    # mm

    def __init__(self, prob_type="discrete"):

        super().__init__(name='TenBarTruss')
        self.prob_type = prob_type
        self.structure = self.create_structure()
        self.create_variables(profiles=TEN_BAR_AREAS_mm2)
        self.create_constraints()
        self.create_objective()


    def create_variables(self, profiles=[0, 1e6]):

        """
        Creates variables used in optimization

        Appends each variable to 'vars' 'list from where they can be accessed
        """
        self.vars = []
        for i, mem in enumerate(self.structure.members.values()):
            name = 'A' + str(i + 1)
            if self.prob_type == "discrete":
                var = DiscreteVariable(name,
                                    values=profiles,
                                    target={"property": "A",
                                            "objects": [mem]})

            elif self.prob_type == "continuous":
                var = Variable(name,
                               lb=profiles[0],
                               ub=profiles[-1],
                               target={"property": "A",
                                       "objects": [mem]})

            elif self.prob_type == 'index':
                var = IndexVariable(name,
                                   values=profiles,
                                   target={"property": "A",
                                           "objects": [mem]})


            else:

                raise TypeError("Problem type must be either 'discrete' "
                                "or 'continuous")


            self.add(var)

    def create_structure(self):

        nodes = []
        nodes.append([2 * L, L])
        nodes.append([2 * L, 0])
        nodes.append([L, L])
        nodes.append([L, 0])
        nodes.append([0, 0])
        nodes.append([0, L])

        mems = [[4, 2], [2, 0], [5, 3], [3, 1], [2, 3], [0, 1], [4, 3], [5, 2], [2, 1], [3, 0]]

        profs = []
        for i in range(10):
            profs.append('SHS 50x50x3.0')

        t = SimpleTruss(nodes, mems, profs)

        t.add(PointLoad(nodes[1], [0, -F, 0]))
        t.add(PointLoad(nodes[3], [0, -F, 0]))

        t.add(XYHingedSupport(nodes[4]))
        t.add(XYHingedSupport(nodes[5]))

        # Change material properties
        for mem in t.members.values():
            mem.material.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = TEN_BAR_AREAS_mm2[0]
            mem.rho = self.rho

        t.generate()
        t.calculate()
        return t

    def constraint_generator(self, mem):

        def compression_fun(x):
            return -mem.ned / (mem.A * mem.fy) - 1

        def tension_fun(x):
            return mem.ned / (mem.A * mem.fy) - 1

        def disp_fun(x):
            displacementsList = list(mem.nodal_displacements.values())
            displacementsMultipleArrays = displacementsList[0].values()

            max_vals = [max(array) for array in displacementsMultipleArrays]
            min_vals = [min(array) for array in displacementsMultipleArrays]

            max_val = max(max_vals)
            min_val = min(min_vals)

            abs_max = max(max_val, abs(min_val))
            return abs_max / self.delta_max - 1

        return compression_fun, tension_fun, disp_fun

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []
        i = 0
        for var in self.vars:
            for mem in var.target["objects"]:
                if isinstance(mem, FrameMember):
                    i += 1
                    compression_fun, tension_fun, disp_fun = self.constraint_generator(mem)

                    comp_con = NonLinearConstraint(con_fun=compression_fun,
                                                   name="Compression " + str(i),
                                                   )
                    comp_con.fea_required = True

                    tension_con = NonLinearConstraint(con_fun=tension_fun,
                                                      name="Tension " + str(i))
                    tension_con.fea_required = True

                    disp_con = NonLinearConstraint(con_fun=disp_fun,
                                                   name='Displacement ' + str(i),
                                                   )
                    disp_con.fea_required = True
                    self.add(comp_con)
                    self.add(tension_con)
                    self.add(disp_con)

    def create_objective(self):

        def objective(X):
            self.substitute_variables(X)
            weight = 0
            for mem in self.structure.members.values():
                weight += mem.weight
            return weight

        obj = ObjectiveFunction(name="Weight",
                                obj_fun=objective)

        self.add(obj)

if __name__ == '__main__':

    from metku.optimization.solvers import ga
    from metku.optimization.solvers import pso
    from metku.optimization.solvers import aco

    problem = SimpleTenBarTruss('discrete')
    solver = pso.PSO(pop_size=100)

    fopt, xopt = solver.solve(problem, maxiter=100)
    print(fopt, xopt)
    problem(xopt)


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 19:04:25 2018

Eurocodes package

Subpackage: en1993

Implementation of EN 1993 Design of steel structures

@author: kmela
"""

from .en1993_1_1 import *
from .en1993_1_2 import *
from .en1993_1_3.en1993_1_3 import *
from .en1993_1_5 import *
from .en1993_1_8.en1993_1_8 import *



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Sun Oct 6 2019

EN 1992-1-1 General rules for concrete structures

@author: kmela
"""

import math

from metku.eurocodes.en1992.constants import gammaC, gammaS, alpha_cc, concrete
from metku.materials.steel_data import Steel


try:
    from metku.eurocodes.en1992.constants import gammaC, gammaS, alpha_cc, concrete
except:
    from eurocodes.en1992.constants import gammaC, gammaS, alpha_cc, concrete
    from materials.steel_data import Steel



def fcd(fck,a_cc=alpha_cc,gC=gammaC):
    """ Design value of compression strength
    """
    return a_cc*fck/gC


class Rebar:
    """ Class for rebar """
    
    def __init__(self,diameter,material="B500B"):
        """ Constructor

            Parameters:
            -----------
                :param diameter: diameter of the rebar (mm)
                :param material: name of the material, to be found in

                Variables:
                ----------
                :ivar values: list of allowable discrete values
                :ivar name: name of the variable (string)
        """
        
        self.d = diameter
        self.mat = Steel(material)
        
        
    @property
    def r(self):
        """ Radius of rebar """
        return 0.5*self.d
    
    @property
    def E(self):
        """ Young's modulus """
        return self.mat.E
    
    @property
    def weight(self):
        """ Weight per unit length (kg/m) """
        return self.mat.rho*self.A*1e3
    
    @property
    def fy(self):
        """ Yield strength """
        return self.mat.fy
    
    @property
    def fyd(self):
        """ Design value of yield strength """
        return self.fy/gammaS
    
    @property
    def fu(self):
        """ Ultimate strength """
        return self.mat.fu
    
    @property
    def A(self):
        """ Cross-sectional area """
        return math.pi*self.d**2/4
    
    @property
    def I(self):
        """ Second moment of area with respect to centroid """
        return 0.25*math.pi*self.r**4


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
def log_to_file(function):
    def wrapper(*args, **kwargs):
        import sys
        import datetime as dt
        import inspect
        import os
        original_stdout = sys.stdout
        filename = f'{function.__name__}_{dt.datetime.now().strftime("%Y_%m_%d__%H_%M_%S")}'
        os.mkdir(f'{os.getcwd()}/{filename}')
        os.chdir(f'{os.getcwd()}/{filename}')
        print(f'writing to file: {filename}')
        with open(f'{filename}.txt', 'w', encoding='UTF-8') as file:
            sys.stdout = file

            print(inspect.getsource(sys.modules['__main__']))
            print('\n\n####################################### SOLUTION ##########################################\n\n')
            function(*args, **kwargs)

            sys.stdout = original_stdout
        os.chdir(f'{os.getcwd()}/..')
        print(f'done writing')
    return wrapper



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  4 16:48:20 2022

Optimization tools for Trusses using Raami framework

@author: kmela
"""

import numpy as np


from metku.eurocodes.en1993.en1993_1_8.rhs_joints import RHSKGapJoint
from metku.raami.raami_plane_truss import TubularKGapJoint
from metku.optimization.structopt import OptimizationProblem
from metku.optimization.variables import Variable
from metku.optimization.constraints import NonLinearConstraint, LinearConstraint
from metku.optimization.objective import ObjectiveFunction
from metku.optimization.solvers.slsqp import SLSQP
from metku.optimization.solvers.trust_region import TrustRegionConstr


def minimize_eccentricity(truss,min_gap=20):
    """ Minimizes eccentrictity of the joints in a tubular truss """
    
    def gap_constraint(joint,min_gap=20):
        """ Generates constraints for the gap in the joint 
            The gap must satisfy:
                0.5*b0*(1-beta) <= gap <= 1.5*b0*(1-beta)
                gap >= t1 + t2
                gap >= min_gap
        """
        beta = joint.joint.beta()
        b0 = joint.joint.b0
        gub = 1.5*b0*(1-beta)
        
        glb = max([0.5*b0*(1-beta),min_gap,sum(joint.t)])
        
        def gap_ub(x):
            """ Returns gap upper bound constraint """
            gap = joint.calc_gap()
            
            return gap - gub
            #return gap/gub-1
        
        def gap_lb(x):
            """ Returns gap lower bound constraint """
            gap = joint.calc_gap()                
            
            return -gap + glb
            #return 1-gap/glb
        
        return gap_ub, gap_lb
    
    def total_eccentricity(opt_vars):
        # Objective function is the sum of eccentricities at K joints.
        C = 100
        yvars = []
        n = len(opt_vars)
        for i, var in enumerate(opt_vars):
            if var.target['property'] == 'yloc':
                yvars.append(i)
        
        def obj_value(x):
            #return 0
            return 1/C*sum(np.array(x)[yvars]**2)
        
        def obj_grad(x):
            """ Gradient of the objective function """
            df = np.zeros(n)
            df[yvars] = 2/C*np.array(x)[yvars]
            
            return df
        
        return obj_value, obj_grad
        
    
    xlb = -200
    xub = 200
    # Create empty optimization problem
    P = OptimizationProblem(structure=truss)
    
    # Create joint variables and constraints
    for key, joint in truss.joints.items():
        # Assume symmetry of the truss, so only need to consider
        # half of the joints.
        if joint.node.x <= 0.5*truss.span:            
            if joint == truss.ridge_joint:
                new_var = Variable(f'y_{key}',lb=-0.55*joint.h0,ub=0.25*joint.h0,value=0.1*joint.h0,
                                   target={'property':'yloc','objects':[joint]})
                P.add(new_var)
                
                # Add constraints for the gap
                gap_ub, gap_lb = gap_constraint(joint,min_gap)
                gap_ub_con = NonLinearConstraint(gap_ub,con_type="<",name=f'Gap upper bound, joint {key}',problem=P)
                gap_lb_con = NonLinearConstraint(gap_lb,con_type="<",name=f'Gap lower bound, joint {key}',problem=P)
                P.add(gap_ub_con)
                P.add(gap_lb_con)
            elif isinstance(joint,TubularKGapJoint):
                
                new_y_var = Variable(f'y_{key}',lb=-0.55*joint.h0,ub=0.25*joint.h0,value=0.1*joint.h0,
                                         target={'property':'yloc','objects':[joint]})
                new_x_var = Variable(f'x_{key}',lb=xlb,ub=xub,value=0,
                                         target={'property':'xloc','objects':[joint]})
                
                P.add(new_y_var)                
                P.add(new_x_var)
                
                # Add constraints for the gap
                gap_ub, gap_lb = gap_constraint(joint,min_gap)
                gap_ub_con = NonLinearConstraint(gap_ub,con_type="<",name=f'Gap upper bound, joint {key}',problem=P)
                gap_lb_con = NonLinearConstraint(gap_lb,con_type="<",name=f'Gap lower bound, joint {key}',problem=P)
                P.add(gap_ub_con)
                P.add(gap_lb_con)
    
    x0 = np.array([var.value for var in P.vars])
    
    # Generate constraints:
    # For vertical members, add constraints that keep the members vertical
    n = len(P.vars)
    for key, brace in truss.braces.items():        
        if brace.vertical:
            
            a = np.zeros(n)
            # We need to identify the joints of the member
            # and the corresponding variables in order to write constraints.
            
            for i, var in enumerate(P.vars):
                if var.target['objects'][0].node == brace.nodes[0]:
                    print("Variable connected to node 1")
                    if var.target['property'] == 'xloc':
                        print("Variable related to local x coordinate")
                        a[i] = var.target['objects'][0].chords[0].dir_vector[0]
                    elif var.target['property'] == 'yloc':
                        print("Variable related to local y coordinate")
                        a[i] = var.target['objects'][0].chords[0].perpendicular[0]
                elif var.target['objects'][0].node == brace.nodes[1]:
                    print("Variable connected to node 2")
                    if var.target['property'] == 'xloc':
                        print("Variable related to local x coordinate")
                        a[i] = -var.target['objects'][0].chords[0].dir_vector[0]
                    elif var.target['property'] == 'yloc':
                        print("Variable related to local y coordinate")
                        a[i] = -var.target['objects'][0].chords[0].perpendicular[0]
            
            if any(a):            
                P.add(LinearConstraint(a,b=0,con_type="=",name=f"Vertical brace {key}"))
            
    #
    
    # Objective function
    obj_fun, obj_grad = total_eccentricity(P.vars)
    P.add(ObjectiveFunction("Eccentricity", obj_fun))
    P.grad = obj_grad
    
    return P, x0

if __name__ == "__main__":
    
    from metku.raami.raami_plane_truss import Ktruss_example, Ntruss_example
    
    #t = Ktruss_example(h2=1800,h1=1500,dx1=1000, dx2=1000,first=True, edges=False)
    t = Ntruss_example(h2=1700,dx1=2000, dx2=2000, first=False,edges=True)
    t.symmetry()
    t.generate_supports()
    t.generate_joints()
    t.determine_symmetry_joints()
    t.generate_uniform_load(q=-25)
    t.generate_fem(model='en1993')
    #t.generate_fem(model='no_eccentricity')
    
    t.optimize_members(verb=True)
    t.plot(geometry=True,loads=False)
    
    P, x0 = minimize_eccentricity(t)
    #x0[0] = 20
    
    solver = SLSQP()
    #solver = TrustRegionConstr()
    min_ecc, xmin = solver.solve(P,x0=x0,verb=True)
    
    t.plot(geometry=True,loads=False)
    
    #P(x0)
    



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Fri May 17 10:42 2019

Welded box sections

@author: kmela
"""
import math
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.patches as patches


from metku.eurocodes.en1993 import en1993_1_1, en1993_1_5, constants
from metku.sections.steel.steel_section import SteelSection



class WBSection(SteelSection):
    """ Welded box sections """

    def __init__(self, h, tw, b, tf, c, fy=355, weld_throat=5):
        """
            Mono-symmetric welded I-section
            
            h -- height of the profile
            b -- width of the flanges
            tw .. thickness of web
            tf .. tf thickness of flanges                
            c .. length of the lip
            fy -- yield strength
            weld_throat .. throat thickness of the weld between
                           flanges and web (fillet weld assumed)
        """
        self.h = h
        self.b = b
        self.tf = tf
        self.tw = tw
        self.weld_throat = weld_throat
        self.lip = c        
        #self.fy = fy
        
        # Reduction factors for effective width of parts
        self.rho_web = 1.0
        self.rho_top = 1.0
        self.rho_bottom = 1.0
        
        A = self.area()
        Au = self.paint_area()
        I = self.second_moment()
        Ashear = self.shear_area(fy)
        Wel = self.section_modulus()
        Wpl = self.plastic_section_modulus()

        self.It = self.torsional_constant()
        self.Iw = self.warping_constant()

        super().__init__(fy, A, I, Au, Wpl, Wel, Ashear)

        """ Determine buckling curve: EN 1993-1-1, Table 6.2 """ 
        if max(tf) <= 40:
            self.imp_factor = [en1993_1_1.buckling_curve["b"],
                               en1993_1_1.buckling_curve["c"]]
        else:
            self.imp_factor = [en1993_1_1.buckling_curve["c"],
                               en1993_1_1.buckling_curve["d"]]

        if h/min(b) <= 2.0:
            self.imp_factor_LT_gen = en1993_1_1.buckling_curve["c"]
            self.imp_factor_LT = en1993_1_1.buckling_curve["c"]
        else:
            self.imp_factor_LT_gen = en1993_1_1.buckling_curve["d"]
            self.imp_factor_LT = en1993_1_1.buckling_curve["d"]
            
    def __getattribute__(self, name):
        """ override the attribute access for those attributes
            that depend on the section dimensions and that are
            constant for rolled sections
        """
        if name == "A":
            if self.area() < 0:
                print(self.h, self.b, self.tt, self.tb, self.tw)
                
            return self.area()
        elif name == "I":
            I = self.second_moment()
            if I[1] < 0:
                print(self.h, self.b, self.tt, self.tb, self.tw)
            return self.second_moment()
        elif name == "Wel":
            return self.section_modulus()
        elif name == "Wpl":
            return self.plastic_section_modulus()
        elif name == "Ashear":
            return self.shear_area()
        elif name == "Au":
            return self.paint_area()
        elif name == "It":
            return self.torsional_constant()
        elif name == "Iw":
            return self.warping_constant()
        elif name == "zs":
            return self.shear_centre()
        elif name == "zj":
            return self.wagners_factor()
        elif name == "b":
            return np.array([self.bt, self.bb])
        elif name == "tf":
            return np.array([self.tt, self.tb])
        else:
            return super().__getattribute__(name)

    def __setattr__(self, key, value):
        if key == "b":        
            self.bt = value
            self.bb = value
        elif key == "tf":
            self.tt = value
            self.tb = value
        #elif key == "bt":
        #    self.b[0] = value
        #elif key == "bb":
        #    self.b[1] = value
        else:
            super().__setattr__(key, value)

    @property
    def hw(self):
        """ Web height """
        return self.h - 2*self.tf
    
    @property
    def Aw(self):
        """ Area of web plate """
        return self.hw * self.tw
    
    @property
    def Izw(self):
        """ Second moment of area of web with respect to its own major axis """
        return self.hw * self.tw**3/12

    @property
    def Iyw(self):
        """ Second moment of area of web with respect to its own major axis """
        return self.tw * self.hw ** 3 / 12
    
    @property
    def zw(self):
        """ Distance of centroid of web from bottom of section """
        return self.tb + 0.5 * self.hw

    # @property
    # def tb(self):
    #     """ Thickness of bottom flange """
    #     return self.tf[1]
    #
    # @property
    # def bb(self):
    #     """ Width of bottom flange """
    #     return self.b[1]
    
    @property
    def Af(self):
        """ Area of one flange """
        return self.b * self.tf
    
    @property
    def Izb(self):
        """ Second moment of area of bottom flange with respect to its own
            major axis
        """
        return self.tb * self.bb**3/12

    @property
    def Iyb(self):
        """ Second moment of area of bottom flange with respect to its own
            major axis
        """
        return self.bb * self.tb ** 3 / 12
    
    @property
    def zb(self):
        """ Distance of centroid of bottom flange from bottom of section """
        return 0.5 * self.tb
        
    @property
    def Izt(self):
        """ Second moment of area of top flange with respect to its own major
            axis
        """
        return self.tt * self.bt**3/12

    @property
    def Iyt(self):
        """ Second moment of area of top flange with respect to its own major
            axis
        """
        return self.bt * self.tt ** 3 / 12
    
    @property
    def zt(self):
        """ Distance of centroid of top flange from bottom of section """
        return self.h - 0.5 * self.tt
    
    @property
    def cf_top(self):
        """ Straight part of the top flange """
        return 0.5*(self.bt-self.tw) - math.sqrt(2)*self.weld_throat

    @property
    def cf_bottom(self):
        """ Straight part of the bottom flange """
        return 0.5*(self.bb-self.tw) - math.sqrt(2)*self.weld_throat

    @property
    def cw(self):
        """ Straight part of the web """
        return self.hw - 2*math.sqrt(2)*self.weld_throat
   
    @property
    def MfRd(self):
        """ Plastic bending resistanc of flanges only
            Commentary and Worked Examples to EN 1993-1-5, pp.70, Fig. 5.11
        """
        
        # Distance between center lines of flanges
        hf = self.h-0.5*(self.tb+self.tt)
        Af1 = self.At
        Af2 = self.Ab
        return min(hf*Af1*self.fy,hf*Af2*self.fy)
    
    @property
    def NtopRd(self):
        """ Axial resistance of top flange
        """
        return self.At*self.fy
    
    @property
    def NbottomRd(self):
        """ Axial resistance of bottom flange
        """
        return self.Ab*self.fy
    
    
    def Aeff(self):
        """ Effective area of the section """
        
    
    def shear_centre(self):
        """ Position of the shear centre S from centroid G
            Source: Maquoi et al. (2003)
            Lateral torsional buckling in steel and composite beams,
            Book 2, 7.1.2 Mono-symmetrical I cross-sections
        """
        
        """ Position of the centroids of the different plates
            with respect to centroid of the cross-section
        """
        zG = self.elastic_neutral_axis()
        zft = self.zt-zG
        zfb = self.zb-zG
        zwS = self.zw-zG
        
        return (self.Izt*zft + self.Izb*zfb + self.Izw*zwS)/self.I[1]

    def wagners_factor(self):
        """ Source: Maquoi et al. (2003)
            Lateral torsional buckling in steel and composite beams,
            Book 2, 7.1.2 Mono-symmetrical I cross-sections
        """

        zf1 = self.zt
        zf2 = self.zb
        zw = self.zw
        Izf1 = self.Izt
        Izf2 = self.Izb
        Izw = self.Izw
        Iyf1 = self.Iyt
        Iyf2 = self.Iyb
        Iyw = self.Iyw
        Af1 = self.At
        Af2 = self.Ab
        Aw = self.Aw

        zs = self.zs

        zj = zs - (1 / (2 * self.I[0])) * (
                zf1 * (Izf1 + Af1 * zf1 ** 2 + 3 * Iyf1)
                + (zf2 * (Izf2 + Af2 * zf2 ** 2 + 3 * Iyf2))
                + (zw * (Izw + Aw * zw ** 2 + 3 * Iyw)))

        return zj
    
    def flange_class(self, verb=False):
        """ Determine class of compressed flange """
        # cf = 0.5*(self.b - self.tw) - self.r
        rf = self.cf_top / self.tt
        cFlange = en1993_1_1.outstand_part_in_compression(rf, self.eps)

        if verb:
            print("Flange classification (outstand element):")
            print("cf = {0:4.2f}, tf = {1:4.2f}".format(self.cf_top, self.tt))
            print("cf/tf = {0:4.2f}".format(rf))
            print("Flange class = {0}".format(cFlange))

        return cFlange

    def web_class_comp(self, verb=False):
        """ Determine class of compressed web """
        rw = self.cw / self.tw
        cWeb = en1993_1_1.internal_part_in_compression(rw, self.eps)

        if verb:
            print("Web classification (internal part in compression):")
            print("cw = {0:4.2f}, tw = {1:4.2f}".format(self.cw, self.tw))
            print("cw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_bend(self, verb=False):
        """ Determine class of web in bending """
        zel = self.elastic_neutral_axis()
        dpl = self.plastic_neutral_axis()
        rw = self.cw / self.tw
        psi = (-zel+self.tb)/(self.h-zel-self.tt)

        if dpl <= self.tb + math.sqrt(2) * self.weld_throat:
            alpha = 1
        elif dpl >= self.h - self.tt - math.sqrt(2) * self.weld_throat:
            alpha = 0  # TODO tarkista tämä
        else:
            alpha = (self.h - self.tt - math.sqrt(2) * self.weld_throat - dpl)\
                    / (self.h - self.tt - self.tb - 2 * math.sqrt(2))

        cWeb = en1993_1_1.internal_part_comp_bend(rw, self.eps, alpha, psi)

        if verb:
            print("Web classification (internal part in compression and "
                  "bending):")
            print("cw = {0:4.2f}, tw = {1:4.2f}".format(self.cw, self.tw))
            print("cw/tw = {0:4.2f}".format(rw))
            print("alpha = {0:4.2f}".format(alpha))
            print("psi = {0:4.2f}".format(psi))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_comp_bend(self, Ned, verb=False):
        """ Determine class of web in combined bending and compression 
            TODO tarkista tämä
        """
        # cw = self.h-2*self.tf-2*self.r
        rw = self.hw / self.tw
        Ac = 0.5 * (self.A + Ned / (self.fy / constants.gammaM0))
        a = Ac / self.A
        p = -1

        cWeb = en1993_1_1.internal_part_comp_bend(rw, self.eps, a, p)

        if verb:
            print("Web classification (internal part in compression and "
                  "bending):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw, self.tw))
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def moment_axial_force_interact(self, UN, MRd):
        """ Interaction rule for section resistance for combined
            axial force and bending
            
            input: UN .. NEd/NRd
                  MRd .. moment resistance           
            TODO tarkista tämä
        """
        aw = min((self.A - (self.bt * self.tt - self.bb * self.tb))
                 / self.A, 0.5)

        if UN > 1.0:
            MNRd = 0.0
        else:
            MNRd = MRd * min((1 - UN) / (1 - 0.5 * aw), 1)

        return MNRd

    def area(self):
        """ Cross-sectional area """
        return self.Aw + 2*self.Af
    
    def shear_area(self,fy=None):
        """ Shear area """
        # if self.hw < 1e-6 or self.tw < 1e-3:
        #     print(self.hw, self.tw)
        if fy is None:
            fy = self.fy
        
        return en1993_1_5.shear_eta(fy)*self.hw*self.tw
    
    def paint_area(self):
        """ Area to be painted (circumference of the section) """
        return 2*self.tt + self.bt + 2*self.hw + 2*self.tb + 2*self.bb \
               - 2*self.tw

    def elastic_neutral_axis(self):
        """ Elastic neutral axis 
            measured from the bottom of the section
        """
        zel = (self.zb*self.Ab + self.zw*self.Aw + self.zt*self.At)/self.area()
        return zel
    
    def plastic_neutral_axis(self):
        """ Plastic neutral axis
            measured from the bottom of the section
        """
        # measured downwards from the bottom of the top flange
        # dpl = 0.5*(self.Ab + self.Aw - self.At)/self.tw

        # with respect to y-axis
        if self.At >= self.A / 2:
            dpl = self.h - self.A / (2 * self.bt)
        elif self.Ab >= self.A / 2:
            dpl = self.A / (2 * self.bb)
        else:
            dpl = (self.A / 2 - self.Ab) / self.tw + self.tb

        #  print(dpl)
        return dpl
    
    def second_moment(self):
        """ Second moment of area with respect to both principal axes"""
        zel = self.elastic_neutral_axis()
                
        # with respect to z axis
        Iy = 1/12*self.bb*self.tb**3 + (zel-self.zb)**2*self.Ab + \
                1/12*self.tw*self.hw**3 + (zel-self.zw)**2*self.Aw + \
                1/12*self.bt*self.tt**3 + (zel-self.zt)**2*self.At

        # with respect to z axis        
        Iz = 1/12*self.tb*self.bb**3 + 1/12*self.tt*self.bt**3 + \
                1/12*self.hw*self.tw**3

        return [Iy, Iz]
    
    def section_modulus(self):
        """ Elastic section modulus for bending """
        
        I = self.second_moment()
        zel = self.elastic_neutral_axis()
        
        # distance of neutral axis from the top of the section
        ztop = self.h - zel
        
        Wel = [0, 0]
        Wel[0] = I[0]/max(zel, ztop)
        
        # with respect to z-axis
        Wel[1] = I[1]/(0.5*max(self.b))
        
        return np.asarray(Wel)
    
    def plastic_section_modulus(self):
        """ Plastic section modulus for bending """
        
        """ distance of plastic neutral axis from the bottom of the top 
        flange """

        Wpl = [0, 0]
    
        # with respect to y-axis
        dpl = self.plastic_neutral_axis()

        if self.At >= self.A / 2:
            # neutral axis is in the top flange
            Wpl[0] = (self.bt * pow(self.h - dpl, 2) / 2) \
                     + (self.bt * pow(dpl - self.hw - self.tb, 2) / 2) \
                     + (self.Aw * (dpl - self.hw/2-self.tb)) \
                     + (self.Ab * (dpl - self.tb / 2))

        elif self.Ab >= self.A / 2:
            # neutral axis is in the bottom flange
            Wpl[0] = (self.At * (self.h - dpl - self.tt / 2)) \
                     + (self.Aw * (self.h - dpl - self.tt - self.hw / 2)) \
                     + (self.bb * pow(self.h - dpl -
                                      self.tt - self.hw, 2) / 2) \
                     + (self.bb * pow(dpl, 2) / 2)

        else:
            # neutral axis is in the web
            Wpl[0] = (self.At * (self.h - dpl - self.tt / 2)) \
                     + (self.tw * pow(self.h - dpl - self.tt, 2) / 2) \
                     + (self.tw * pow(dpl - self.tb, 2) / 2) \
                     + (self.Ab * (dpl - self.tb / 2))
            
        # with respect to z-axis
        Wpl[1] = (1/4 * self.Ab * self.bb) + (1/4 * self.At * self.bt) \
                 + (1/4 * self.Aw * self.tw)

        return np.asarray(Wpl)

    def torsional_constant(self):
        """ Torsional constant """

        It = 1/3 * (self.bt * pow(self.tt, 3) + self.bb *
                    pow(self.tb, 3) + self.hw * pow(self.tw, 3))

        # Tarkempi kaava (hidastaa laskentaa) ja lisätermit lähellä arvoa 1
        # It = 1/3 * (self.bt * pow(self.tt, 3) * (1 - 0.63 * (
        #         self.tt / self.bt) * (1 - pow(self.tt, 4) / (12 * pow(
        #             self.bt, 4)))) + self.bb * pow(self.tb, 3) * (1 - 0.63 * (
        #                 self.tb / self.bb) * (1 - pow(self.tb, 4) / (12 * pow(
        #                     self.bb, 4)))) + self.hw * pow(self.tw, 3))

        return It

    def warping_constant(self):
        """ Warping constant on the weaker axis """

        Izf1 = (self.tt * pow(self.bt, 3)) / 12
        Iw = Izf1 * (1 - Izf1 / self.I[1]) * \
             pow(self.h - self.tt / 2 - self.tb / 2, 2)

        return Iw

    def effective_flange(self, flange="Top"):
        """ Effective flange, assumed to be in pure compression """
        
        ksigma = en1993_1_5.buckling_factor_outstand(psi=1.0)
                
        if flange == "Top":
            c = self.cf_top
            t = self.tt
        else:
            c = self.cf_bottom
            t = self.tb

        lp = en1993_1_5.lambda_p(c, t, self.eps, ksigma)
        rho = en1993_1_5.reduction_factor_outstand(lp)

        return rho
    
    def effective_web(self,psi=None,verb=False):
        """ Effective web of the cross-section """
        
        cw = self.cw
        
        """ If stress ratio is not provided, calculate it """
        if psi == None:
            zel = self.elastic_neutral_axis()
            psi = (-zel+self.tb)/(self.h-zel-self.tt)
        
        ksigma = en1993_1_5.buckling_factor_internal(psi)
        lp = en1993_1_5.lambda_p(cw, self.tw, self.eps, ksigma)
        rho = en1993_1_5.reduction_factor_internal(lp,psi)
        
        beff, be = en1993_1_5.effective_width_internal(cw,psi,rho)
        
        if verb:
            print("Effective Web:")
            print("psi = {0:4.2f}".format(psi))
            print("k_sigma = {0:4.3f}".format(ksigma))
            print("lambda_p = {0:4.3f}".format(lp))
            print("rho = {0:4.3f}".format(rho))
            print("beff = {0:4.3f}".format(beff))
            print("be1 = {0:4.3f}, be2 = {1:4.3f} ".format(be[0],be[1]))
        
        

    
    def shear_buckling_resistance(self,stiffener_spacing=10e5,end_post="non-rigid"):
        """ Shear buckling resistance according to EN 1993-1-5, Clause 5 
            input:
                stiffener_spacing .. spacing between adjacent stiffeners
                end_post .. "non-rigid" or "rigid"
            
        """
    
        hw = self.hw
        tw = self.tw
        b = self.bt
        a = stiffener_spacing
        rw = hw/tw
        fyw = self.fy
        eta = en1993_1_5.shear_eta(fyw)
    
        if rw < 72*self.eps/eta:
            """ No need for shear buckling:
                return shear resistance of section
            """
            VbRd = self.VRd
        else:           
            """ Contribution from the web """
            tau_cr = en1993_1_5.tau_crit(hw,a,tw,b)            
            slend_w = en1993_1_5.shear_buckling_slenderness(fyw,tau_cr)
            chi_w = en1993_1_5.shear_buckling_reduction_factor(slend_w,eta,end_post)
            Vbw_Rd = en1993_1_5.shear_buckling_web(chi_w,fyw,hw,t)
            
            """ Contribution from the flange """
            NEd = abs(self.Ned)
            
            """ If axial force is present, the resistance MfRd must
                be reduced by the factor (1-rN). (EN 1993-1-5, 5.4(2))
            """
            if NEd > 1e-6:
                rN = NEd/((self.At+self.Ab)*fy)
            else:
                rN = 0
            
            rM = abs(self.Med)/((1-rN)*self.MfRd)
            
            if rM < 1.0:
                if self.NtopRd <= self.NbottomRd:
                    bf = self.bt
                    tf = self.tt
                else:
                    bf = self.bb
                    tf = self.tb
                
                bf = min(bf,30*e*tf+tw)
                
                Vbf_Rd = en1993_1_5.shear_buckling_flanges(bf,tf,fy,a,hw,t,fy,rM)
            
    
    def draw(self, name=""):
        """ Draw the profile """
        fig, ax = plt.subplots(1)
        
        # create section parts
        bot_flange = patches.Rectangle((-0.5*self.bb, 0), width=self.bb,
                                       height=self.tb, fill=False, hatch='\\')
        web = patches.Rectangle((-0.5*self.tw, self.tb), width=self.tw,
                                height=self.hw, fill=False, hatch='\\')
    
        top_flange = patches.Rectangle((-0.5*self.bt, self.h-self.tt),
                                       width=self.bt, height=self.tt,
                                       fill=False, hatch='\\')
        ax.add_patch(bot_flange)
        ax.add_patch(web)
        ax.add_patch(top_flange)
        ax.set_xlim(-0.5*max(self.b), 0.5*max(self.b))
        ax.set_ylim(0, self.h)
        
        zel = self.elastic_neutral_axis()
        dpl = self.plastic_neutral_axis()        
        
        print(zel)
        print(dpl)
        
        plt.plot(0.0, zel, 'or')
        plt.plot(0.0, dpl, 'db')
        
        ax.set_aspect('equal')
        
        #  plt.show()

        plt.savefig('testi' + 'plot.png')


def test_sym():
    """ symmetric I-beam """
    p = WISection(500, 8, [250, 250], [12, 12])
    print(p.MRd)
    print("Done.")
    # p.draw()


def test_non_sym():
    """ symmetric I-beam """
    print("Create section")
    p = WISection(500, 8, [300, 250], [10, 15])
    print(p.MRd)
    print("Done.")
    # p.draw()


if __name__ == '__main__':

    from copy import deepcopy
    
    #test_non_sym()
    #from frame2d.frame2d import *
    #frame = Frame2D()
    #col = SteelColumn([[0, 0], [0, 5000]])
    #frame.add(col)
    
    #p = WISection(1594, 7.0, [300.0, 300.0], [22.0, 22.0],fy=235,weld_throat=3.5)    
    p = WISection(1000, 6.0, [300.0, 300.0], [15.0, 15.0],fy=355,weld_throat=4)    
    p.Med = 2.228e6    
    p.section_class(verb=True)
    p.effective_web(verb=True)
    p.draw()
    
    #s = p
    
    s = WISection(p.h,p.tw,p.b,p.tf,p.fy,p.weld_throat)
    s.h = 1200
    
    print(s.h)
    print(p.h)
    
    #profile = "WI 200-5-200-8-100-5"
    # frame.plot()
    #print(col.cross_section.b)
    #print(col.cross_section.tf)
    #col.cross_section.b = 150
    #col.cross_section.tf = 6
    #print(col.cross_section.b)
    #print(col.cross_section.tf)
    # col.cross_section.draw()




# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 15:14:17 2022

@author: kmela
"""

import numpy as np
from collections.abc import Iterable

from metku.optimization.constants import X_TOL, POT_ACT_TOL
from metku.optimization.variables import Variable, IndexVariable
import metku.optimization.opt_functions as opt_funs


class Constraint:
    """ General class for constraints """

    def __init__(self, name="", con_type="<",
                 problem=None, fea_required=False,
                 vars=None,grad=None):
        """ Constructor
            Input:
                name: Name of the constraint (string)
                con_type: '<', '>' or '='
                problem: OptimizationProblem class object, to which the
                         constraint belongs
                fea_required: (Boolean) to indicate whether or not carry
                              out structural analysis, when the constraint
                              is evaluated
                vars: list of Variable type objects
                grad: callable function that returns the gradient of constraint function
        """
        self.name = name
        self.type = con_type
        self.problem = problem
        self.fea_required = fea_required
        if not isinstance(vars, Iterable):
            vars = [vars]
        self.vars = vars
        
        # Lagrange multiplier
        self.mult = 1.0
        
        self.grad = grad
        self.potential = True # Flag for stating whether the constraint is potentially active or not
        self.pot_tol = POT_ACT_TOL

    def __call__(self, x):
        """ Evaluate the constraint using the syntax con(x)
            
        Runs finite element analysis on problem's structure if needed
        """
        if self.problem is not None:
            """ Collect current variable values stored in each
                optimization variable to X
            """
            
            X = np.array([var.value for var in self.problem.vars])            
            x = np.asarray(x)
            
            """ If the new values deviate from the old values
                in terms of L2 norm by more than X_TOL, substitute
                new values.

                NOTE: X_TOL should be small
            """
            if np.linalg.norm(X - x) > X_TOL:
                #print('variablet asetettiin')
                self.problem.substitute_variables(x)

        """ If structural analysis is part of the consraint evaluation
            AND
            if FEM analysis has not been carried out for X, perform FEM
        """
        if self.fea_required and not self.problem.fea_done:            
            self.problem.fea()

    def __repr__(self):
        return f"{type(self).__name__}: {self.name}"

    def neg_call(self, x):
        """ Evaluate -g(x) """
        return -self(x)


class LinearConstraint(Constraint):
    """ Class for linear constraints """

    def __init__(self, a=None, b=None, con_type="<", name="", problem=None, **kwargs):
        """ Constructor 
            Constraint is of the form: a'*x con_type b
            
            **kwargs: can be 'fea_required' and 'vars'.
            Probably fea_required is always False for linear constraints
        """

        self.a = np.array(a)
        self.b = b
        Constraint.__init__(self, name, con_type, problem, **kwargs)

    def __call__(self, x):
        """ Evaluate constraint at x 
        """
        
        """ This call substitutes variable values (if needed) and
            performs structural analysis (if needed)
        """
        super().__call__(x)

        X = []
        """ By iterating over 'all_vars', also the fixed variables
            are included in the function evaluation.
        """
        for var in self.problem.all_vars:
            if isinstance(var, IndexVariable):
                X.append(var.idx)
            else:
                X.append(var.value)

        if isinstance(self.b, Variable):
            b = self.b.value
        else:
            b = self.b

        gval = self.a.dot(np.array(X)) - b
        
        if self.type == '<':
            if gval < -self.pot_tol:
                self.potential = False
            else:
                self.potential = True
        elif self.type == '>':
            if gval > self.pot_tol:
                self.potential = False
            else:
                self.potential = True

        return gval
        


class NonLinearConstraint(Constraint):
    """ Class for linear constraints """

    def __init__(self, con_fun, con_type="<", name="", problem=None, **kwargs):
        """ Constructor
            con_fun: function that returns the value of constraint function
                     g(x) <= 0 (or >= 0 or = 0)
        """

        self.con = con_fun
        self.grad = None        
        Constraint.__init__(self, name, con_type, problem, **kwargs)

    def __call__(self, x):
        """ Evaluate constraint at x """
        
        """ This call substitutes variable values (if needed) and
            performs structural analysis (if needed)
        """        
        super().__call__(x)

        
        """ Make vector 'fixed_vals', corresponding to the vector of
            design variables that includes those in 'x' and the fixed
            values
        """
        X = [val for val in x]
        X.reverse()
        
        """ fixed_vals is a list with length equalling the
            number of original variable, including fixed and free.
            In place of free variables, 'fixed_vals' has None, and to this
            position, the value from 'X' should be inserted.
        """
        # IS COPY ENOUGH, OR SHOULD IT BE deepcopy?
        fixed_vals = self.problem.fixed_vals.copy() 
        for i, val in enumerate(fixed_vals):
            if val is None:
                fixed_vals[i] = X.pop()
            if not X:
                break

        gval = self.con(fixed_vals)    

        if self.type == '<':
            if gval < -self.pot_tol:
                self.potential = False
            else:
                self.potential = True
        elif self.type == '>':
            if gval > self.pot_tol:
                self.potential = False
            else:
                self.potential = True

        return gval
    
    def df(self,x):
        """ Evaluate gradient at x """
        
        if self.grad is None:
            # Evaluate gradient numerically
            df = opt_funs.NumGrad(self.con,x)
        else:
            df = self.grad(x)
        
        return df